"""
Abstractions for defining closures.

A closure is a a set of physical equations which determines the sub-mesh
turbulence of the water column at a specific time. The purpose of Tunax is to
calibrate the parameters of these equations. This module contains the abstract
classes required to define a closure, which is done in the folder
:code:`closures/`. These classes can be obtained by the prefix
:code:`tunax.closure.` or directly by :code:`tunax.`.

"""

from abc import ABC
from typing import Callable, Type, TypeVar, Generic

import equinox as eqx
from jaxtyping import Float, Array

from tunax.case import Case
from tunax.space import Grid, State


class ClosureParametersAbstract(eqx.Module, ABC):
    """
    Abstraction for the parameters of a closure.
    
    To define a closure, a class that inherits from this one must be created.
    This parent class does not impose anything on the child one. The child
    class should be fill of the parameters involved in the closure which may be
    calibrated. The child class must be the only place where these parameters
    are defined for running a forward model : the instance of this class is
    given to the closure functions to recover the parameters and do the
    computation. For a run of a calibration, the description of the parameters
    to calibration is done from the child class. It can also includes
    parameters that are not dedicated to be calibrated such as physical or
    mathematical constants. The attributes of this class can be used to avoid
    the systematic computation of some values which are independent of the time
    (eg. k-epsilon with the method :code:`__post_init__`). 
    
    """


# variable that represent a type which contains ClosureParametersAbstract and
# all its subclasses
CloParT = TypeVar('CloParT', bound=ClosureParametersAbstract)


class ClosureStateAbstract(eqx.Module, ABC):
    r"""
    Abstraction for the water column state linked to the closure.

    To define a closure, a class that inherits rom this one must be created.
    This parent class imposes a grid and three variables as attributes : these
    variables are used at each step of the forward model to compute the next
    step of tracers and momentum. The child class should be fill with other
    arrays and scalars that may evolve throught time and necessary for the
    computation of the closure at the next step. The child class is similary
    to the :class:`~space.State` class but for the diffusivity part.
    The :code:`__init__` method can be overwrited to fill the variables with
    initial values.

    Parameters
    ----------
    grid : Grid
        cf. attribute.
    akt : Float[~jax.Array, 'nz+1']
        cf. attribute.
    akv : Float[~jax.Array, 'nz+1']
        cf. attribute.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column, should be the same than for the
        :class:`~space.State` instance used in the model.
    akt : Float[~jax.Array, 'nz+1']
        Eddy-diffusivity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : Float[~jax.Array, 'nz+1']
        Eddy-viscosity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.

    """

    grid: Grid
    akt: Float[Array, 'nz+1']
    akv: Float[Array, 'nz+1']


# variable that represent a type which contains ClosureStateAbstract and
# all its subclasses
CloStateT = TypeVar('CloStateT', bound=ClosureStateAbstract)


class Closure(eqx.Module, Generic[CloStateT, CloParT]):
    r"""
    Implementation of a physical closure for computing eddy-diffusivity.

    This class contains the three parts that defines a closure of a vertical
    physics. These parts will be used in the model to compute the eddy-
    diffusivity and eddy-viscosity at each time-step from the current water
    column :class:`~space.State`, the current water column state of the
    closure (child class of :class:`~ClosureStateAbstract`), the physical
    :class:`~case.Case` and the closure parameters (child class of
    :class:`~ClosureParametersAbstract`).

    Parameters
    ----------
    name : str
        cf. attribute.
    parameters_class : Type[ClosureParametersAbstract]
        cf. attribute.
    state_class : Type[ClosureStateAbstract]
        cf. attribute.
    step_fun : Callable[[State, CloStateT, float, CloParT, Case], CloStateT]
        cf. attribute.

    Attributes
    ----------
    name : str
        The name of the closure.
    parameters_class : Type[ClosureParametersAbstract]
        A child class of :class:`~ClosureParametersAbstract` that defines the
        constant parameters used in the computation done by the closure, it
        includes the parameters that may be calibrated.
    state_class : Type[ClosureStateAbstract]
        A child class of :class:`~ClosureStateAbstract` that defines the state
        of the water column for the variables used by the closure computation.
    step_fun : Callable[[State, CloStateT, float, CloParT, Case], CloStateT]
        Ths function is called at every step of the forward model to compute
        the eddy-diffusivity before resolving the equation of the tracers and
        of the momentum.
        
        Parameters
        ----------
        state : State
            Current state of the water column.
        closure_state : CloStateT
            Current state of the water column for the variables used by the
            closure.
        dt : float
            Time-step of the forward model :math:`[\text s]`.
        closure_params : CloParT
            Values of the parameters used by the closure (time-independant).
        case : Case
            Physical parameters and forcings of the model run.
        
        Returns
        -------
        closure_state : CloStateT
            State of the water column for the variables used by the
            closure at the next time-step.
        
        Notes
        -----
        :code:`CloStateT` is the type reprensenting the instances of the child
        classes of :class:`~ClosureStateAbstract` and :code:`CloParT` is the
        same for :class:`~ClosureParametersAbstract`.

    """

    name : str
    parameters_class: Type[ClosureParametersAbstract]
    state_class: Type[ClosureStateAbstract]
    step_fun: Callable[[State, CloStateT, float, CloParT, Case], CloStateT]



r"""
:math:`k-\varepsilon` closure parameters, states and computation functions.

This module contains the implementation of the :math:`k-\varepsilon` model
described as a GLS case as in [1]_ as a :class:`~closure.Closure` instance. The
model was traduced from Frotran to JAX with the work of Florian Lemarié and
Manolis Perrot [2]_, the translation was done in part using the work of Anthony
Zhou, Linnia Hawkins and Pierre Gentine [3]_. The parameters of the closure are
available in the :class:`KepsParameters` class, the closure state in
:class:`KepsState` class. The function :attr:`keps_step` compute one time-step
of the closure, which means that it computes the eddy-diffusivity and
viscosity. The module contains other functions that are used by this main one.
These classes and the function step can be obtained by the prefix
:code:`tunax.closures.k_epsilon` or directly by :code:`tunax.closures`.

References
----------
.. [1] L. Umlauf and H. Burchard. A generic length-scale equation for
    geophysical turbulence models (2003). Journal of Marine Research 61
    pp. 235-265. doi : `10.1357/002224003322005087 <https://www.semanticscholar
    .org/paper/A-generic-length-scale-equation-for-geophysical-Umlauf-Burchard/
    24fd6403615fc7a6c5d9b6156e4f1e8d4d280af2>`_.
.. [2] M. Perrot and F. Lemarié. Energetically consistent Eddy-Diffusivity
    Mass-Flux convective schemes. Part I: Theory and Models (2024). url :
    `hal.science/hal-04439113 <https://hal.science/hal-04439113>`_.
.. [3] A. Zhou, L. Hawkins and P. Gentine. Proof-of-concept: Using ChatGPT to
    Translate and Modernize an Earth System Model from Fortran to Python/JAX
    (2024). url : `arxiv.org/abs/2405.00018
    <https://arxiv.org/abs/2405.00018>`_.
    
"""


from __future__ import annotations
from functools import partial
from typing import Tuple

import equinox as eqx
import jax.numpy as jnp
from jax import lax, jit
from jaxtyping import Float, Array

from tunax.case import Case
from tunax.space import Grid, State
from tunax.functions import tridiag_solve, add_boundaries
from tunax.closure import ClosureParametersAbstract, ClosureStateAbstract


class KepsParameters(ClosureParametersAbstract):
    r"""
    Parameters and constants for :math:`k-\varepsilon`.

    The first 17 attributes are the parameters of :math:`k-\varepsilon` that
    may be calibrated. This class also contains some physical constants used
    in the closure computing. The last 19 attributes are the one for the
    stability function that are computed from the parameters of
    :math:`k-\varepsilon`.

    Parameters
    ----------
    c1 : float, default=5.
        cf. attribute.
    c2 : float, default=0.8
        cf. attribute.
    c3 : float, default=1.968
        cf. attribute.
    c4 : float, default=1.136
        cf. attribute.
    c5 : float, default=0.
        cf. attribute.
    c6 : float, default=0.4
        cf. attribute.
    cb1 : float, default=5.95
        cf. attribute.
    cb2 : float, default=.6
        cf. attribute.
    cb3 : float, default=1.
        cf. attribute.
    cb4 : float, default=0.
        cf. attribute.
    cb5 : float, default=0.3333
        cf. attribute.
    cbb : float, default=.72
        cf. attribute.
    c_mu0 : float, default=0.5477
        cf. attribute.
    sig_k : float, default=1.
        cf. attribute.
    sig_eps : float, default=1.3
        cf. attribute.
    c_eps1 : float, default=1.44
        cf. attribute.
    c_eps2 : float, default=1.92
        cf. attribute.
    c_eps3m : float, default=-0.4
        cf. attribute.
    c_eps3p : float, default=1.
        cf. attribute.
    chk_grav : float, default=1400.
        cf. attribute.
    galp: float, default=0.53
        cf. attribute.
    z0s_min : float, default=1e-2
        cf. attribute.
    z0b_min : float, default=1e-4
        cf. attribute.
    z0b : float, default=1e-14
        cf. attribute.
    akt_min : float, default=1e-5
        cf. attribute.
    akt_min : float, default=1e-4
        cf. attribute.
    tke_min : float, default=1e-6
        cf. attribute.
    eps_min : float, default=1e-12
        cf. attribute.
    c_mu_min : float, default=0.1
        cf. attribute.
    c_mu_prim_min : float, default=0.1
        cf. attribute.
    dir_sfc: bool, default=False
        cf. attribute.
    dir_btm: bool, default=True
        cf. attribute.
    gls_p : float, default=3
        cf. attribute.
    gls_m : float, default=1.5
        cf. attribute.
    gls_n : float, default=-1
        cf. attribute.

    Attributes
    ----------
    c1 : float, default=5.
        :math:`k-\varepsilon` parameter :math:`c_1` for the dissipation of the
        corelation tensor pressure/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c2 : float, default=0.8
        :math:`k-\varepsilon` parameter :math:`c_2` for the dissipation of the
        corelation tensor pressure/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c3 : float, default=1.968
        :math:`k-\varepsilon` parameter :math:`c_3` for the dissipation of the
        corelation tensor pressure/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c4 : float, default=1.136
        :math:`k-\varepsilon` parameter :math:`c_4` for the dissipation of the
        corelation tensor pressure/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c5 : float, default=0.
        :math:`k-\varepsilon` parameter :math:`c_5` for the dissipation of the
        corelation tensor pressure/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c6 : float, default=0.4
        :math:`k-\varepsilon` parameter :math:`c_6` for the dissipation of the
        corelation tensor pressure/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    cb1 : float, default=5.95
        :math:`k-\varepsilon` parameter :math:`c_{b1}` for the dissipation of
        the corelation tensor buoyancy/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    cb2 : float, default=.6
        :math:`k-\varepsilon` parameter :math:`c_{b2}` for the dissipation of
        the corelation tensor buoyancy/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    cb3 : float, default=1.
        :math:`k-\varepsilon` parameter :math:`c_{b3}` for the dissipation of
        the corelation tensor buoyancy/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    cb4 : float, default=0.
        :math:`k-\varepsilon` parameter :math:`c_{b4}` for the dissipation of
        the corelation tensor buoyancy/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    cb5 : float, default=0.3333
        :math:`k-\varepsilon` parameter :math:`c_{b5}` for the dissipation of
        the corelation tensor buoyancy/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    cbb : float, default=.72
        :math:`k-\varepsilon` parameter :math:`c_{b}` for the dissipation of
        the corelation tensor buoyancy/velocity (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c_mu0 : float, default=0.5477
        :math:`k-\varepsilon` parameter :math:`c_\mu^0` which links the mixing
        length to the dissipation (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    sig_k : float, default=1.
        :math:`k-\varepsilon` parameter :math:`\sigma_k` Schmit number for the
        dissipation of TKE (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    sig_eps : float, default=1.3
        :math:`k-\varepsilon` parameter :math:`\sigma_\varepsilon` Schmit
        number for the dissipation of :math:`\varepsilon` (Umlauf and Burchard
        notations) :math:`[\text{dimensionless}]`.
    c_eps1 : float, default=1.44
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 1}` correction of
        the :math:`\varepsilon` equation (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c_eps2 : float, default=1.92
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 2}` correction of
        the :math:`\varepsilon` equation (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c_eps3m : float, default=-0.4
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 3}^-` correction
        of the :math:`\varepsilon` equation (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    c_eps3p : float, default=1.
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 3}^+` correction
        of the :math:`\varepsilon` equation (Umlauf and Burchard notations)
        :math:`[\text{dimensionless}]`.
    chk_grav : float, default=1400.
        Charnock coefficient times gravity :math:`[\text{dimensionless}]`.
    galp: float, default=0.53
        Parameter for Galperin mixing length limitation
        :math:`[\text{dimensionless}]`.
    z0s_min : float, default=1e-2
        Minimal surface roughness length :math:`[\text m]`.
    z0b_min : float, default=1e-4
        Minimal bottom roughness length :math:`[\text m]`.
    z0b : float, default=1e-14
        Bottom roughness length :math:`[\text m]`.
    akt_min : float, default=1e-5
        Minimal and initialization value of eddy-diffusivity
        :math:`\left[\text m^2 \cdot \text s^{-1} \right]`.
    akv_min : float, default=1e-4
        Minimal and initialization value of eddy-viscosity
        :math:`\left[\text m^2 \cdot \text s^{-1} \right]`.
    tke_min : float, default=1e-6
        Minimal and initialization value of turbulent kinetic energy (TKE)
        :math:`\left[\text m^3 \cdot \text s^{-2} \right]`.
    eps_min : float, default=1e-12
        Minimal and initialization value of TKE dissipation
        :math:`\left[\text m^2 \cdot \text s^{-3} \right]`.
    c_mu_min : float, default=0.1
        Minimal and initialization value of :math:`c_\mu` in GLS formalisim
        :math:`[\text{dimensionless}]`.
    c_mu_prim_min : float, default=0.1
        Minimal and initialization value of `c_\mu'` in GLS formalisim
        :math:`[\text{dimensionless}]`.
    dir_sfc: bool, default=False
        Apply a Dirichlet boundary condition at the surface for TKE, else
        apply a Neumann boundary condition.
    dir_btm: bool, default=True
        Apply a Dirichlet boundary condition at the bottom for TKE, else
        apply a Neumann boundary condition.
    gls_p : float, default=3
        GLS coefficient :math:`p` to define :math:`k-\varepsilon`
        :math:`[\text{dimensionless}]`.
    gls_m : float, default=1.5
        GLS coefficient :math:`m` to define :math:`k-\varepsilon`
        :math:`[\text{dimensionless}]`.
    gls_n : float, default=-1
        GLS coefficient :math:`n` to define :math:`k-\varepsilon`
        :math:`[\text{dimensionless}]`.
    sf_d0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_d1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_d2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_d3 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_d4 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_d5 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_n0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_n1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_n2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_nb0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_nb1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    sf_nb2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    lim_am0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    lim_am1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    lim_am2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    lim_am3 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    lim_am4 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    lim_am5 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    lim_am6 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the
        parameters :math:`[\text{dimensionless}]`.
    
    """

    # k-epsilon coefficients (Umlauf and Burchard notations)
    c1: float = 5.
    c2: float = .8
    c3: float = 1.968
    c4: float = 1.136
    c5: float = 0.
    c6: float = .4
    cb1: float = 5.95
    cb2: float = .6
    cb3: float = 1.
    cb4: float = 0.
    cb5: float = 0.3333
    cbb: float = 0.72
    c_mu0: float = 0.5477
    sig_k: float = 1.
    sig_eps: float = 1.3
    c_eps1: float = 1.44
    c_eps2: float = 1.92
    c_eps3m: float = -.4
    c_eps3p: float = 1.
    # physical constants
    chk_grav: float = 1400.
    galp: float = .53
    z0s_min: float = 1e-2
    z0b_min: float = 1e-2
    z0b: float = 1e-14
    akt_min: float = 1e-5
    akv_min: float = 1e-4
    tke_min: float = 1e-6
    eps_min: float = 1e-12
    c_mu_min: float = .1
    c_mu_prim_min: float = .1
    # physical case
    dir_sfc: bool = False
    dir_btm: bool = True
    # GLS coefficient for k-epsilon
    gls_p: float = 3
    gls_m: float = 1.5
    gls_n: float = -1
    # limitation coefficients computed from k-epsilon coefficients
    sf_d0: float = eqx.field(init=False)
    sf_d1: float = eqx.field(init=False)
    sf_d2: float = eqx.field(init=False)
    sf_d3: float = eqx.field(init=False)
    sf_d4: float = eqx.field(init=False)
    sf_d5: float = eqx.field(init=False)
    sf_n0: float = eqx.field(init=False)
    sf_n1: float = eqx.field(init=False)
    sf_n2: float = eqx.field(init=False)
    sf_nb0: float = eqx.field(init=False)
    sf_nb1: float = eqx.field(init=False)
    sf_nb2: float = eqx.field(init=False)
    lim_am0: float = eqx.field(init=False)
    lim_am1: float = eqx.field(init=False)
    lim_am2: float = eqx.field(init=False)
    lim_am3: float = eqx.field(init=False)
    lim_am4: float = eqx.field(init=False)
    lim_am5: float = eqx.field(init=False)
    lim_am6: float = eqx.field(init=False)

    def __post_init__(self):
        # stability function coefficients
        a1 = .66666666667 - .5*self.c2
        a2 = 1 - .5*self.c3
        a3 = 1 - .5*self.c4
        a5 = .5 - .5*self.c6
        nn = .5*self.c1
        nb = self.cb1
        ab1 = 1 - self.cb2
        ab2 = 1 - self.cb3
        ab3 = 2*(1 - self.cb4)
        ab5 = 2*self.cbb*(1-self.cb5)
        sf_d0 = 36*nn*nn*nn*nb*nb
        sf_d1 = 84*a5*ab3*nn*nn*nb + 36*ab5*nn*nn*nn*nb
        sf_d2 = 9*(ab2*ab2 - ab1*ab1)*nn*nn*nn - 12*(a2*a2 - 3*a3*a3)*nn*nb*nb
        sf_d3 = 12*a5*ab3*(a2*ab1 - 3*a3*ab2)*nn + \
            12*a5*ab3*(a3*a3 - a2*a2)*nb + 12*ab5*(3*a3*a3 - a2*a2)*nn*nb
        sf_d4 = 48*a5*a5*ab3*ab3*nn + 36*a5*ab3*ab5*nn*nn
        sf_d5 = 3*(a2*a2 - 3*a3*a3)*(ab1*ab1 - ab2*ab2)*nn
        sf_n0 = 36*a1*nn*nn*nb*nb
        sf_n1 = -12*a5*ab3*(ab1 + ab2)*nn*nn + \
            8*a5*ab3*(6*a1 - a2 - 3*a3)*nn*nb + 36*a1*ab5*nn*nn*nb
        sf_n2 = 9*a1*(ab2*ab2 - ab1*ab1)*nn*nn
        sf_nb0 = 12*ab3*nn*nn*nn*nb
        sf_nb1 = 12*a5*ab3*ab3*nn*nn
        sf_nb2 = 9*a1*ab3*(ab1 - ab2)*nn*nn + \
            (6*a1*(a2 - 3*a3) - 4*(a2*a2 - 3*a3*a3))*ab3*nn*nb
        lim_am0 = sf_d0*sf_n0
        lim_am1 = sf_d0*sf_n1 + sf_d1*sf_n0
        lim_am2 = sf_d1*sf_n1 + sf_d4*sf_n0
        lim_am3 = sf_d4*sf_n1
        lim_am4 = sf_d2*sf_n0
        lim_am5 = sf_d2*sf_n1 + sf_d3*sf_n0
        lim_am6 = sf_d3*sf_n1
        object.__setattr__(self, 'sf_d0', sf_d0)
        object.__setattr__(self, 'sf_d1', sf_d1)
        object.__setattr__(self, 'sf_d2', sf_d2)
        object.__setattr__(self, 'sf_d3', sf_d3)
        object.__setattr__(self, 'sf_d4', sf_d4)
        object.__setattr__(self, 'sf_d5', sf_d5)
        object.__setattr__(self, 'sf_n0', sf_n0)
        object.__setattr__(self, 'sf_n1', sf_n1)
        object.__setattr__(self, 'sf_n2', sf_n2)
        object.__setattr__(self, 'sf_nb0', sf_nb0)
        object.__setattr__(self, 'sf_nb1', sf_nb1)
        object.__setattr__(self, 'sf_nb2', sf_nb2)
        object.__setattr__(self, 'lim_am0', lim_am0)
        object.__setattr__(self, 'lim_am1', lim_am1)
        object.__setattr__(self, 'lim_am2', lim_am2)
        object.__setattr__(self, 'lim_am3', lim_am3)
        object.__setattr__(self, 'lim_am4', lim_am4)
        object.__setattr__(self, 'lim_am5', lim_am5)
        object.__setattr__(self, 'lim_am6', lim_am6)


class KepsState(ClosureStateAbstract):
    r"""
    Define the state of the water column for the :math:`k-\varepsilon` model.

    The first initilisation is done from the minimal values of the different
    variables given in an instance of :class:'KepsParameters`.

    Parameters
    ----------
    grid : Grid
        cf. attribute.
    keps_params : KepsParameters
        Used to define the initialization values of the variables.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column, should be the same than for the
        :class:`~space.State` instance used in the model.
    akt : Float[~jax.Array, 'nz+1']
        Eddy-diffusivity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : Float[~jax.Array, 'nz+1']
        Eddy-viscosity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    tke :  Float[~jax.Array, 'nz+1']
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of
        the cells :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    eps :  Float[~jax.Array, 'nz+1']
        TKE dissipation denoted :math:`\varepsilon` on the interfaces of the
        cells :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    c_mu :  Float[~jax.Array, 'nz+1']
        :math:`c_\mu` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.
    c_mu_prim :  Float[~jax.Array, 'nz+1']
        :math:`c_\mu'` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.

    """

    grid: Grid
    akt: Float[Array, 'nz+1']
    akv: Float[Array, 'nz+1']
    tke: Float[Array, 'nz+1']
    eps: Float[Array, 'nz+1']
    c_mu: Float[Array, 'nz+1']
    c_mu_prim: Float[Array, 'nz+1']

    def __init__(self, grid: Grid, keps_params: KepsParameters):
        self.grid = grid
        nz = grid.nz
        self.akt = jnp.full(nz+1, keps_params.akt_min)
        self.akv = jnp.full(nz+1, keps_params.akv_min)
        self.tke = jnp.full(nz+1, keps_params.tke_min)
        self.eps = jnp.full(nz+1, keps_params.eps_min)
        self.c_mu = jnp.full(nz+1, keps_params.c_mu_min)
        self.c_mu_prim = jnp.full(nz+1, keps_params.c_mu_prim_min)


@partial(jit, static_argnames=('case',))
def keps_step(
        state: State,
        keps_state: KepsState,
        dt: float,
        keps_params: KepsParameters,
        case: Case
    ) -> KepsState:
    r"""
    Run one time-step of the :math:`k-\varepsilon` closure.

    The purpose of this function is to get the eddy-diffusivity and
    eddy-viscosity at the next time-step. It works in 3 steps

    1. The Brunt–Väisälä frequency and the shear is computed from the
       :code:`state` and the boundary conditions are computed.
    2. The equations on :math:`k` and :math:`\varepsilon` are solved and their
       values are computed for the next time step.
    3. The eddy-diffusivity and viscosity are computed as diagnostic variables
       and the :math:`keps_state` is updated.

    Parameters
    ----------
    state : State
        Current state of the water column.
    keps_state : KepsState
        Current state of the water column for the variables used by
        :math:`k-\varepsilon`.
    dt : float
        Time-step of the forward model :math:`[\text s]`.
    keps_params: KepsParameters
        Values of the parameters used by :math:`k-\varepsilon` (time-
        independant).
    case : Case
        Physical parameters and forcings of the model run.
    
    Returns
    -------
    keps_state : KepsState
            State of the water column for the variables used by
            :math:`k-\varepsilon` at the next time-step.

    Note
    ----
    This function is jitted with JAX, it should make it faster, but the
    :func:`~jax.jit` decorator can be removed.
    """
    akt = keps_state.akt
    akv = keps_state.akv
    tke = keps_state.tke
    eps = keps_state.eps
    c_mu = keps_state.c_mu
    c_mu_prim = keps_state.c_mu_prim
    u = state.u
    v = state.v
    zr = state.grid.zr
    hz = state.grid.hz

    # prognostic computations
    _, bvf = compute_rho_eos(state.t, state.s, zr, case)
    shear2 = compute_shear(u, v, u, v, zr)
    tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc = compute_tke_eps_bc(
        tke, hz, keps_params, case
    )

    # integrations
    tke_new = advance_turb(
        akt, akv, tke, tke, eps, c_mu, c_mu_prim, bvf, shear2, hz, dt,
        tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc, keps_params, True
    )
    eps_new = advance_turb(
        akt, akv, tke, tke_new, eps, c_mu, c_mu_prim, bvf, shear2, hz, dt,
        tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc, keps_params, False
    )

    # diagnostic variables
    akt_new, akv_new, eps_new, c_mu_new, c_mu_prim_new = compute_diag(
        tke_new, eps_new, bvf, shear2, keps_params
    )

    keps_state = eqx.tree_at(lambda t: t.akv, keps_state, akv_new)
    keps_state = eqx.tree_at(lambda t: t.akt, keps_state, akt_new)
    keps_state = eqx.tree_at(lambda t: t.tke, keps_state, tke_new)
    keps_state = eqx.tree_at(lambda t: t.eps, keps_state, eps_new)
    keps_state = eqx.tree_at(lambda t: t.c_mu, keps_state, c_mu_new)
    keps_state = eqx.tree_at(lambda t: t.c_mu_prim, keps_state, c_mu_prim_new)

    return keps_state


def compute_rho_eos(
        t: Float[Array, 'nz'],
        s: Float[Array, 'nz'],
        zr: Float[Array, 'nz'],
        case: Case
    ) -> Tuple[Float[Array, 'nz+1'], Float[Array, 'nz']]:
    r"""
    Compute density anomaly and Brunt–Väisälä frequency.
    
    Prognostic computation via linear Equation Of State (EOS) :

    :math:`\rho = \rho_0(1-\alpha (T-T_0) + \beta (S-S_0))`

    :math:`N^2 = - \dfrac g {\rho_0} \partial_z \rho`

    Parameters
    ----------
    t : Float[~jax.Array, 'nz']
        Temperature on the center of the cells :math:`[° \text C]`.
    s : Float[~jax.Array, 'nz']
        Salinity on the center of the cells :math:`[\text{psu}]`.
    zr : Float[~jax.Array, 'nz']
        Depths of cell centers from deepest to shallowest :math:`[\text m]`
    case : Case
        Physical parameters and forcings of the model run.

    Returns
    -------
    bvf : Float[Array, 'nz+1']
        Brunt–Väisälä frequency squared :math:`N^2` on cell interfaces
        :math:`\left[\text s^{-2}\right]`.
    rho : Float[Array, 'nz']
        Density anomaly :math:`\rho` on cell interfaces 
        :math:`\left[\text {kg} \cdot \text m^{-3}\right]`
    """
    rho0 = case.rho0
    rho = rho0 * (1. - case.alpha*(t-case.t_rho_ref) + \
                  case.beta*(s-case.s_rho_ref))
    cff = 1./(zr[1:]-zr[:-1])
    bvf_in = - cff*case.grav/rho0 * (rho[1:]-rho[:-1])
    bvf = add_boundaries(0., bvf_in, bvf_in[-1])

    return rho, bvf


def compute_shear(
        u_n: Float[Array, 'nz'],
        v_n: Float[Array, 'nz'],
        u_np1: Float[Array, 'nz'],
        v_np1: Float[Array, 'nz'],
        zr: Float[Array, 'nz']
    ) -> Float[Array, 'nz+1']:
    r"""
    Compute shear production term for TKE equation.

    The prognostic equations are

    :math:`S_h^2 = \partial_Z U^n \cdot \partial_z U^{n+1/2}`

    where :math:`U^{n+1/2}` is the mean between :math:`U^n` and
    :math:`U^{n+1}`.
    
    Parameters
    ----------
    u_n : Float[~jax.Array, 'nz']
        Current zonal velocity on the center of the cells
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    v_n : Float[~jax.Array, 'nz']
        Current meridional velocity on the center of the cells
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    u_np1 : Float[~jax.Array, 'nz']
        Zonal velocity on the center of the cells at the next time step
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    v_np1 : Float[~jax.Array, 'nz']
        Meridional velocity on the center of the cells at the next time step
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    zr : Float[~jax.Array, 'nz']
        Depths of cell centers from deepest to shallowest :math:`[\text m]`
        

    Returns
    -------
    shear2 : Float[~jax.Array, 'nz+1']
        Shear production squared :math:`S_h^2` on cell interfaces
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    """
    cff = 1.0 / (zr[1:] - zr[:-1])**2
    du = 0.5*cff * (u_np1[1:]-u_np1[:-1]) * \
        (u_n[1:]+u_np1[1:]-u_n[:-1]-u_np1[:-1])
    dv = 0.5*cff * (v_np1[1:]-v_np1[:-1]) * \
        (v_n[1:]+v_np1[1:]-v_n[:-1]-v_np1[:-1])
    shear2_in = du + dv
    return add_boundaries(0., shear2_in, 0.)


def compute_tke_eps_bc(
        tke: Float[Array, 'nz+1'],
        hz: Float[Array, 'nz'],
        keps_params: KepsParameters,
        case: Case
    ) -> Tuple[float, float, float, float]:
    r"""
    Compute top and bottom boundary conditions for TKE and :math:`\varepsilon`.

    Parameters
    ----------
    tke : Float[~jax.Array, 'nz+1']
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of
        the cells :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    hz : Float[~jax.Array, 'nz']
        Thickness of cells from deepest to shallowest :math:`[\text m]`.
    keps_params : KepsParameters
        Values of the parameters used by :math:`k-\varepsilon`.
    case : Case
        Physical parameters and forcings of the model run.

    Returns
    -------
    tke_sfc_bc : float
        TKE value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    tke_btm_bc : float
        TKE value for bottom boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    eps_sfc_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).
    eps_btm_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).

    Note
    ----
    The kind of boundary conditions between Neumann and Dirichlet are register
    in the parameters :code:`keps_params`.
    """
    # constants
    rp, rm, rn = keps_params.gls_p, keps_params.gls_m, keps_params.gls_n
    c_mu0 = keps_params.c_mu0
    cm0inv2 = 1./c_mu0**2
    vkarmn = case.vkarmn
    chk = keps_params.chk_grav/case.grav
    sig_eps = keps_params.sig_eps

    # velocity scales
    ustar2_sfc = jnp.sqrt(case.ustr_sfc**2 + case.vstr_sfc**2)
    ustar2_bot = jnp.sqrt(case.ustr_btm**2 + case.vstr_btm**2)

    # TKE Dirichlet boundary condition
    tke_sfc_dir = jnp.maximum(keps_params.tke_min, cm0inv2*ustar2_sfc)
    tke_btm_dir = jnp.maximum(keps_params.tke_min, cm0inv2*ustar2_bot)

    # TKE Neumann boundary condition
    tke_sfc_neu = 0.0
    tke_btm_neu = 0.0

    # epsilon surface conditions
    z0_s = jnp.maximum(keps_params.z0s_min, chk*ustar2_sfc)
    lgthsc = vkarmn*(0.5*hz[-1] + z0_s)
    tke_sfc = 0.5*(tke[-1]+tke[-2])
    eps_sfc_dir = jnp.maximum(keps_params.eps_min, \
        c_mu0**rp * lgthsc**rn * tke_sfc**rm)
    eps_sfc_neu = -rn*vkarmn/sig_eps * c_mu0**(rp+1) * \
        tke_sfc**(rm+.5) * lgthsc**rn

    # epsilon bottom conditions
    z0b = jnp.maximum(keps_params.z0b, keps_params.z0b_min)
    lgthsc = vkarmn *(0.5*hz[0] + z0b)
    tke_btm = 0.5*(tke[0]+tke[1])
    eps_btm_dir = jnp.maximum(keps_params.eps_min, \
        c_mu0**rp * lgthsc**rn * tke_btm**rm)
    eps_btm_neu = -rn*vkarmn/sig_eps * c_mu0**(rp+1) * \
        tke_btm**(rm+.5) * lgthsc**rn

    tke_sfc_bc = jnp.where(keps_params.dir_sfc, tke_sfc_dir, tke_sfc_neu)
    tke_btm_bc = jnp.where(keps_params.dir_btm, tke_btm_dir, tke_btm_neu)
    eps_sfc_bc = jnp.where(keps_params.dir_sfc, eps_sfc_dir, eps_sfc_neu)
    eps_btm_bc = jnp.where(keps_params.dir_btm, eps_btm_dir, eps_btm_neu)

    return tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc


def advance_turb(
        akt: Float[Array, 'nz+1'],
        akv: Float[Array, 'nz+1'],
        tke: Float[Array, 'nz+1'],
        tke_np1: Float[Array, 'nz+1'],
        eps: Float[Array, 'nz+1'],
        c_mu: Float[Array, 'nz+1'],
        c_mu_prim: Float[Array, 'nz+1'],
        bvf: Float[Array, 'nz+1'],
        shear2: Float[Array, 'nz+1'],
        hz: Float[Array, 'nz'],
        dt: float,
        tke_sfc_bc: float,
        tke_btm_bc: float,
        eps_sfc_bc: float,
        eps_btm_bc: float,
        keps_params: KepsParameters,
        do_tke: bool
    ) -> Float[Array, 'nz+1']:
    r"""
    Integrate TKE or :math:`\varepsilon` quantities.

    First the shear and buoyancy production are computed, then they are used in
    the building of the tridiagonal problem, the boundary conditions are then
    added and finally the tridiagonal problem is solved.

    Parameters
    ----------
    akt : Float[~jax.Array, 'nz+1']
        Current eddy-diffusivity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : Float[~jax.Array, 'nz+1']
        Current eddy-viscosity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    tke : Float[~jax.Array, 'nz+1']
        Current turbulent kinetic energy (TKE) denoted :math:`k` on the
        interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    tke_np1 : Float[~jax.Array, 'nz+1']
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of
        the cells at next step (usefull only for :math:`\varepsilon`
        integration) :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    eps : Float[~jax.Array, 'nz+1']
        Current TKE dissipation denoted :math:`\varepsilon` on the interfaces
        of the cells :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    c_mu : Float[~jax.Array, 'nz+1']
        Current :math:`c_\mu` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.
    c_mu_prim : Float[~jax.Array, 'nz+1']
        Current :math:`c_\mu'` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.
    bvf : float(nz+1)
        Current Brunt–Väisälä frequency squared :math:`N^2` on cell interfaces
        :math:`\left[\text s^{-2}\right]`.
    shear2 : Float[~jax.Array, 'nz+1']
        Current shear production squared :math:`S_h^2` on cell interfaces
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    hz : Float[~jax.Array, 'nz']
        Thickness of cells from deepest to shallowest :math:`[\text m]`.
    dt : float
        Time-step of the forward model :math:`[\text s]`.
    tke_sfc_bc : float
        TKE value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    tke_btm_bc : float
        TKE value for bottom boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    eps_sfc_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).
    eps_btm_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).
    keps_params : KepsParameters
        Values of the parameters used by :math:`k-\varepsilon`.
    do_tke : bool
        If :code:`True` solve the equation for TKE, else for
        :math:`\varepsilon`.

    Returns
    -------
    vec : Float[~jax.Array, 'nz+1']
        TKE or :math:`\varepsilon` at next step (depending on :code:`do_tke`).
    """
    # fill the matrix off-diagonal terms for the tridiagonal problem
    cff = -0.5*dt
    ak_vec = jnp.where(do_tke, akv/keps_params.sig_k, akv/keps_params.sig_eps)
    a_in = cff*(ak_vec[1:-1]+ak_vec[:-2]) / hz[:-1]
    c_in = cff*(ak_vec[1:-1]+ak_vec[2:]) / hz[1:]

    # shear and buoyancy production
    s_prod_tke = akv[1:-1]*shear2[1:-1]
    b_prod_tke = -akt[1:-1]*bvf[1:-1]
    s_prod_eps = keps_params.c_eps1*c_mu[1:-1]*tke[1:-1]*shear2[1:-1]
    b_prod_eps = -c_mu_prim[1:-1] * tke[1:-1] * \
        (keps_params.c_eps3m*jnp.maximum(bvf[1:-1], 0) + \
         keps_params.c_eps3p*jnp.minimum(bvf[1:-1], 0))
    s_prod = jnp.where(do_tke, s_prod_tke, s_prod_eps)
    b_prod = jnp.where(do_tke, b_prod_tke, b_prod_eps)

    # diagonal and f term
    cff = 0.5*(hz[:-1] + hz[1:])
    f_tke_in = lax.select(b_prod+s_prod > 0,
                            cff*(tke[1:-1]+dt*(b_prod+s_prod)),
                            cff*(tke[1:-1]+dt*s_prod))
    f_eps_in = lax.select(b_prod+s_prod > 0,
                            cff*(eps[1:-1]+dt*(b_prod+s_prod)),
                            cff*(eps[1:-1]+dt*s_prod))
    f_in = jnp.where(do_tke, f_tke_in, f_eps_in)
    b_tke_in = lax.select((b_prod + s_prod) > 0,
        cff*(1. + dt*eps[1:-1]/tke[1:-1]) - a_in - c_in,
        cff*(1. + dt*(eps[1:-1] - b_prod)/tke[1:-1]) - a_in - c_in)
    b_eps_in = lax.select((b_prod + s_prod) > 0,
        cff*(1. + dt*keps_params.c_eps2*eps[1:-1]/tke_np1[1:-1]) - a_in - c_in,
        cff*(1. + dt*keps_params.c_eps2*eps[1:-1]/tke_np1[1:-1] - \
             dt*b_prod/eps[1:-1]) - a_in - c_in)
    b_in = jnp.where(do_tke, b_tke_in, b_eps_in)

    # surface boundary condition
    dir_sfc = keps_params.dir_sfc
    a_sfc = jnp.where(dir_sfc, 0., -0.5*(ak_vec[-1] + ak_vec[-2]))
    b_sfc = jnp.where(dir_sfc, 1., 0.5*(ak_vec[-1] + ak_vec[-2]))
    sfc_bc = jnp.where(do_tke, tke_sfc_bc, eps_sfc_bc)
    f_sfc = jnp.where(dir_sfc, sfc_bc, hz[-1]*sfc_bc)

    # bottom boundary condition
    dir_btm = keps_params.dir_btm
    b_btm = jnp.where(dir_sfc, 1., -0.5*(ak_vec[0] + ak_vec[1]))
    c_btm = jnp.where(dir_sfc, 0., 0.5*(ak_vec[0] + ak_vec[1]))
    btm_bc = jnp.where(do_tke, tke_btm_bc, eps_btm_bc)
    f_btm = jnp.where(dir_btm, btm_bc, hz[0]*btm_bc)

    # vectors rassembly
    a = add_boundaries(0., a_in, a_sfc)
    b = add_boundaries(b_btm, b_in, b_sfc)
    c = add_boundaries(c_btm, c_in, 0.)
    f = add_boundaries(f_btm, f_in, f_sfc)

    # solve tridiagonal problem
    f = tridiag_solve(a, b, c, f)

    vec_min = jnp.where(do_tke, keps_params.tke_min, keps_params.eps_min)
    vec = jnp.maximum(f, vec_min)

    return vec


def compute_diag(
        tke: Float[Array, 'nz+1'],
        eps: Float[Array, 'nz+1'],
        bvf: Float[Array, 'nz+1'],
        shear2: Float[Array, 'nz+1'],
        keps_params: KepsParameters
    ) -> Tuple[Float[Array, 'nz+1'], Float[Array, 'nz+1'],
               Float[Array, 'nz+1'], Float[Array, 'nz+1'],
               Float[Array, 'nz+1']]:
    r"""
    Computes the diagnostic variables of :math:`k-\varepsilon` closure.

    This function first apply the Galperin limitation, then it computes
    :math:`c_\mu'` and :math:`c_\mu` with the stability function, and finally
    it computes the eddy-diffusivity and viscosity with these variables.

    Parameters
    ----------
    tke : Float[~jax.Array, 'nz+1']
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of
        the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    eps : Float[~jax.Array, 'nz+1']
        TKE dissipation denoted :math:`\varepsilon` on the interfaces of the
        cells at next step :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    bvf : Float[~jax.Array, 'nz+1']
        Current Brunt–Väisälä frequency squared :math:`N^2` on cell interfaces
        :math:`\left[\text s^{-2}\right]`.
    shear2 : Float[~jax.Array, 'nz+1']
        Current shear production squared :math:`S_h^2` on cell interfaces
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    keps_params : KepsParameters
        Values of the parameters used by :math:`k-\varepsilon`.

    Returns
    -------
    akt : Float[Array, 'nz+1']
        Eddy-diffusivity on the interfaces of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : Float[Array, 'nz+1']
        Eddy-viscosity on the interfaces of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    eps : Float[Array, 'nz+1']
        TKE dissipation denoted :math:`\varepsilon` on the interfaces
        of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    c_mu : Float[Array, 'nz+1']
        :math:`c_\mu` in GLS formalisim on the interfaces of the cells at next
        step :math:`[\text{dimensionless}]`.
    c_mu_prim : Float[Array, 'nz+1']
        :math:`c_\mu'` in GLS formalisim on the interfaces of the cells at next
        step :math:`[\text{dimensionless}]`.
    """
    # parameters
    akv_min = keps_params.akv_min
    akt_min = keps_params.akt_min
    eps_min = keps_params.eps_min
    sf_d0 = keps_params.sf_d0
    sf_d1 = keps_params.sf_d1
    sf_d2 = keps_params.sf_d2
    sf_d3 = keps_params.sf_d3
    sf_d4 = keps_params.sf_d4
    sf_d5 = keps_params.sf_d5
    sf_n0 = keps_params.sf_n0
    sf_n1 = keps_params.sf_n1
    sf_n2 = keps_params.sf_n2
    sf_nb0 = keps_params.sf_nb0
    sf_nb1 = keps_params.sf_nb1
    sf_nb2 = keps_params.sf_nb2
    lim_am0 = keps_params.lim_am0
    lim_am1 = keps_params.lim_am1
    lim_am2 = keps_params.lim_am2
    lim_am3 = keps_params.lim_am3
    lim_am4 = keps_params.lim_am4
    lim_am5 = keps_params.lim_am5
    lim_am6 = keps_params.lim_am6
    c_mu0 = keps_params.c_mu0
    rp, rm, rn = keps_params.gls_p, keps_params.gls_m, keps_params.gls_n
    e1 = 3 + rp/rn
    e2 = 1.5 + rm/rn
    e3 = -1/rn

    # minimum value of alpha_n to ensure that alpha_m is positive
    alpha_n_min = 0.5*(- (sf_d1 + sf_nb0) + jnp.sqrt((sf_d1 + sf_nb0)**2 - \
        4.0*sf_d0*(sf_d4 + sf_nb1))) / (sf_d4 + sf_nb1)

    # Galperin limitation : l <= l_li
    l_lim = keps_params.galp*jnp.sqrt(2.0*tke[1:-1] / \
        jnp.maximum(1e-14, bvf[1:-1]))

    # limitation (use MAX because rn is negative)
    cff = c_mu0**rp * l_lim**rn * tke[1:-1]**rm
    eps = eps.at[1:-1].set(jnp.maximum(eps[1:-1], cff))
    epsilon = c_mu0**e1 * tke[1:-1]**e2 * eps[1:-1]**e3
    epsilon = jnp.maximum(epsilon, eps_min)

    # compute alpha_n and alpha_m
    cff = (tke[1:-1] / epsilon)**2
    alpha_m = cff*shear2[1:-1]
    alpha_n = cff*bvf[1:-1]

    # limitation of alpha_n and alpha_m
    alpha_n = jnp.minimum(jnp.maximum(0.73*alpha_n_min, alpha_n), 1e10)
    alpha_m_max = (lim_am0 + lim_am1*alpha_n + lim_am2*alpha_n**2 + \
        lim_am3*alpha_n**3) / (lim_am4 + lim_am5*alpha_n + \
        lim_am6*alpha_n**2)
    alpha_m = jnp.minimum(alpha_m, alpha_m_max)

    # compute stability functions
    denom = sf_d0 + sf_d1*alpha_n + sf_d2*alpha_m + sf_d3*alpha_n*alpha_m \
        + sf_d4*alpha_n**2 + sf_d5*alpha_m**2
    cff = 1./denom
    c_mu_in = cff*(sf_n0 + sf_n1*alpha_n + sf_n2*alpha_m)
    c_mu = add_boundaries(keps_params.c_mu_min, c_mu_in, keps_params.c_mu_min)
    c_mu_prim_in = cff*(sf_nb0 + sf_nb1*alpha_n + sf_nb2*alpha_m)
    c_mu_prim = add_boundaries(
        keps_params.c_mu_prim_min, c_mu_prim_in, keps_params.c_mu_prim_min
    )
    epsilon = c_mu0**e1 * tke[1:-1]**e2 * eps[1:-1]**e3
    epsilon = jnp.maximum(epsilon, eps_min)

    # finalize the computation of akv and akt
    cff = tke[1:-1]**2 / epsilon
    akt_in = jnp.maximum(cff*c_mu_prim[1:-1], akt_min)
    akv_in = jnp.maximum(cff*c_mu[1:-1], akv_min)

    akt_btm = jnp.maximum(1.5*akt_in[0] - 0.5*akt_in[1], akt_min)
    akt_sfc = jnp.maximum(1.5*akt_in[-1] - 0.5*akt_in[-2], akt_min)
    akv_btm = jnp.maximum(1.5*akv_in[0] - 0.5*akv_in[1], akv_min)
    akv_sfc = jnp.maximum(1.5*akv_in[-1] - 0.5*akv_in[-2], akv_min)

    akt = add_boundaries(akt_btm, akt_in, akt_sfc)
    akv = add_boundaries(akv_btm, akv_in, akv_sfc)

    return akt, akv, eps, c_mu, c_mu_prim



"""
Importation of Tunax classes and functions for shortcuts.
"""


from .space import Grid, State, Trajectory
from .case import Case
from .closure import (
    ClosureParametersAbstract, ClosureStateAbstract, Closure
)
from .database import Obs, Database
from .functions import tridiag_solve, add_boundaries
from .closures_registry import CLOSURES_REGISTRY
from .model import (
    SingleColumnModel, step, lmd_swfrac, advance_tra_ed, advance_dyn_cor_ed,
    diffusion_solver
)
from .fitter import FittableParameter, FittableParametersSet, Fitter



"""
Registry of available closures.

This module only contains a constant variable which lists all the available
closures. It can be obtained by the prefix :code:`tunax.closures_registry.` or
directly by :code:`tunax.`.

Attributes
==========
CLOSURES_REGISTRY : Dict[str, Closure]
    This variable is a dictionnary whose keys are the name of the closures
    (they must be exaclty the same than :attr:`~closure.Closure.name`), and
    whose values are the corresponding :class:`~closure.Closure` instance of
    the closure. The closure can be used by the forward model and the fitter
    thanks to this constant. When the user add a the code for a closure in
    :code:`closures/` they must

    1. import here the parameters and state classes (child of
       :class:`~closure.ClosureParametersAbstract` and of
       :class:`~closure.ClosureStateAbstract`) and the step function of the
       closure,
    2. from these objects, create the :class:`~closure.Closure` instance of the
       closure,
    3. add an entry at this dictionnary

    The current available closures are :
    
    - k-epsilon

"""

from typing import Dict

from tunax.closure import Closure

from tunax.closures.k_epsilon import KepsParameters, KepsState, keps_step


CLOSURES_REGISTRY: Dict[str, Closure] = {
    'k-epsilon': Closure('k-epsilon', KepsParameters, KepsState, keps_step)
}



"""
Usefull calculation functions.

Thefunctions in this module are supposed to be used in various other modules.
They can be called by the prefix :code:`tunax.functions.` or directly by
:code:`tunax.`.

"""

import jax.numpy as jnp
from jax import lax, jit
from jaxtyping import Float, Array


@jit
def tridiag_solve(
        a: Float[Array, 'n'],
        b: Float[Array, 'n'],
        c: Float[Array, 'n'],
        f: Float[Array, 'n']
    ) -> Float[Array, 'n']:
    r"""
    Solve a trigiagonal problem.

    The tridiagonal problem can be written :math:`\mathbb MX = F` where
    :math:`\mathbb M = \begin{pmatrix} b_1 & c_1 & & \\
    a_2 & \ddots & \ddots & \\
    & \ddots & \ddots & c_{n-1} \\
    & & a_n & b_n
    \end{pmatrix}`
    and :math:`F = \begin{pmatrix} f_1 \\ \vdots \\ f_n \end{pmatrix}`.
    The problem is solved by recurrence using :mod:`jax.lax` 


    Parameters
    ----------
    a : Float[~jax.Array, 'n']
        Left diagonal of :math:`\mathbb M`, the first element is not used.
    b : Float[~jax.Array, 'n']
        Middle diagonal of :math:`\mathbb M`.
    c : Float[~jax.Array, 'n']
        Right diagonal of :math:`\mathbb M`, the last element is not used.
    f : Float[~jax.Array, 'n']
        Right hand of the equation :math:`F`.
    
    Returns
    -------
    x : Float[~jax.Array, 'nz']
        Solution :math:`X` of tridiagonal problem.
    """
    n, = a.shape
    # forward sweep
    cff = 1.0 / b[0]
    f = f.at[0].multiply(cff)
    q = jnp.zeros(n)
    q = q.at[0].set(-c[0] * cff)

    def body_fun1(k: int, x: Float[Array, 'n']):
        f = x[0, :]
        q = x[1, :]
        cff = 1.0 / (b[k] + a[k] * q[k-1])
        q = q.at[k].set(-cff * c[k])
        f = f.at[k].set(cff * (f[k] - a[k] * f[k-1]))
        return jnp.stack([f, q])
    f_q = jnp.stack([f, q])
    f_q = lax.fori_loop(1, n, body_fun1, f_q)
    f = f_q[0, :]
    q = f_q[1, :]

    # backward substitution
    def body_fun2(k: int, x: Float[Array, 'n']):
        return x.at[n-1-k].add(q[n-1-k] * x[n-k])
    x = lax.fori_loop(1, n, body_fun2, f)

    return x


def add_boundaries(
        vec_btm: float,
        vec_in: Float[Array, 'n-2'],
        vec_sfc: float
    ) -> Float[Array, 'n']:
    """
    Concatenate the three parts of a vector : surface, bottom and inside.

    This functions is made to avoid loops and make JAX more efficient by
    writing the calculations with vectorization when possible.

    Parameters
    ----------
    vec_btm : float
        Bottom value of the vector.
    vec_in : Float[~jax.Array, 'n-2']
        Middle values of the vector.
    vec_sfc : float
        Surface value of the vector.

    Returns
    -------

    vec : Float[~jax.Array, 'n']
        Concatenated vector.
    """
    return jnp.concat([jnp.array([vec_btm]), vec_in, jnp.array([vec_sfc])])


def _format_to_single_line(text: str) -> str:
    """
    Transforms a multiple line text in a line string by removing indentations.

    In the code the error and warning messages are written on multiple lines
    with indentations to respect the PEP8 maximum line length of 79 characters
    and the consistency of indentations. This function is used to show
    correctly these messages on one line and without the indentations.

    Parameters
    ----------
    text : str
        Text on multiple lines to transform.

    Returns
    -------
    line : str
        Text on a single line removed from indentations.
    """
    lines = text.splitlines()
    stripped_lines = [line.strip() for line in lines]
    single_line = " ".join(stripped_lines)
    return " ".join(single_line.split())



"""
Single column forward model core.

This module contains the main class :class:`SingleColumnModel` which is the
core of Tunax for implementing computing the evolution of a water column of the
ocean. It also contains functions that are used in this computation. The model
was traduced from Frotran to JAX with the work of Florian Lemarié and Manolis
Perrot [1]_, the translation was done in part using the work of Anthony Zhou,
Linnia Hawkins and Pierre Gentine [2]_. this class and these functions can be
obtained by the prefix :code:`tunax.model.` or directly by :code:`tunax.`.

References
----------
.. [1] M. Perrot and F. Lemarié. Energetically consistent Eddy-Diffusivity
    Mass-Flux convective schemes. Part I: Theory and Models (2024). url :
    `hal.science/hal-04439113 <https://hal.science/hal-04439113>`_.
.. [2] A. Zhou, L. Hawkins and P. Gentine. Proof-of-concept: Using ChatGPT to
    Translate and Modernize an Earth System Model from Fortran to Python/JAX
    (2024). url : `arxiv.org/abs/2405.00018
    <https://arxiv.org/abs/2405.00018>`_.

"""

from __future__ import annotations
import warnings
from typing import Tuple, List
from functools import partial

import equinox as eqx
import jax.numpy as jnp
from jax import lax, jit
from jaxtyping import Float, Array

from tunax.case import Case
from tunax.space import State, Trajectory
from tunax.functions import (
    tridiag_solve, add_boundaries, _format_to_single_line
)
from tunax.closure import (
    ClosureParametersAbstract, ClosureStateAbstract, Closure
)
from tunax.closures_registry import CLOSURES_REGISTRY


class SingleColumnModel(eqx.Module):
    r"""
    Single column forward model core.

    This forward model of tunax is the combination of 4 things : the physical
    case :attr:`case`, an initial state of the water column
    :attr:`init_state`, the time information with :attr:`time_frame`,
    :attr:`dt` and :attr:`out_dt` and the abstraction of the chosen closure
    for eddy-diffusivity :attr:`closure`. Adding a set of parameters for the
    closure one can run the model with the method
    :meth:`compute_trajectory_with`.

    Parameters
    ----------
    time_frame : float
        Total time of the simulation :math:`[\text h]`.
    dt : float
        Time-step of integration for every iteration :math:`[\text s]`.
    out_dt : float
        Time-step for the output writing :math:`[\text s]`.
    init_state : State
        cf. attribute.
    case : Case
        cf. attribute.
    closure_name : str
        Name of the chosen closure, must be a key of
        :attr:`~closures_registry.CLOSURES_REGISTRY`, see its documentation
        for the available closures.

    Attributes
    ----------
    nt : int
        Number of integration interations.
    dt : float
        Time-step of integration for every iteration :math:`[\text s]`.
    n_out : int
        Number of of time-steps between every output.
    init_state : State
        Initial physical state of the water column.
    case : Case
        Physical case and forcings of the experiment.
    closure : Closure
        Abstraction representing the chosen closure.
    
    Warnings
    --------
    - If :code:`time_frame` is not proportional to the time-step :attr:`dt`.
    - If :code:`time_frame` is not proportional to the out time-step
      :code:`out_dt`.
    
    Raises
    ------
    ValueError
        If :code:`out_dt` is not proportional to the time step :attr:`dt`.
    ValueError
        If :code:`closure_name` is not registerd in
        :attr:`~closures_registry.CLOSURES_REGISTRY`.

    Note
    ----
    To make this forward model compatible with the fitter part of Tunax, the
    parameters of the closure are only given during of the call of the run
    with :meth:`compute_trajectory_with`.
        
    """

    nt: int
    dt: float
    n_out: int
    init_state: State
    case: Case
    closure: Closure

    def __init__(
            self,
            time_frame: float,
            dt: float,
            out_dt: float,
            init_state: State,
            case: Case,
            closure_name: str
        ) -> SingleColumnModel:
        # time parameters transformation
        n_out = out_dt/dt
        nt = time_frame*3600/dt

        # warnings and errors on time parameters coherence
        if not n_out.is_integer():
            raise ValueError('`out_dt` should be a multiple of `dt`.')
        if not nt % n_out == 0:
            warnings.warn(_format_to_single_line("""
                The `time_frame`is not proportional to the out time-step
                `out_dt`, the last step will be computed a few before the
                `time_frame`.
            """))
        if not nt.is_integer():
            warnings.warn(_format_to_single_line("""
                The `time_frame`is not proportional to the time-step `dt`, the
                last step will be computed a few before the time_frame.
            """))
        if not closure_name in CLOSURES_REGISTRY:
            raise ValueError(_format_to_single_line("""
                `closure_name` not registerd in CLOSURES_REGISTRY.
            """))

        # write attributes
        self.nt = int(nt)
        self.dt = dt
        self.n_out = int(n_out)
        self.init_state = init_state
        self.case = case
        self.closure = CLOSURES_REGISTRY[closure_name]

    def compute_trajectory_with(
            self,
            closure_parameters: ClosureParametersAbstract
        ) -> Trajectory:
        """
        Run the model with a specific set of closure parameters.

        This method is the main one for runing the model. It calls :attr:`nt`
        times the function :func:`step` and regulary writes the output to build
        the :class:`~space.Trajectory` output.

        Parameters
        ----------
        closure_parameters : ClosureParametersAbstract
            A set of parameters of the used closure.

        Returns
        -------
        trajectory : Trajectory
            Timeseries of the evolution of the variables of the model every
            :code:`out_dt`.
        """
        # initialize the model
        states_list: List[State] = []
        state = self.init_state
        closure_state = self.closure.state_class(
            self.init_state.grid, closure_parameters
        )
        swr_frac = lmd_swfrac(self.init_state.grid.hz)

        # loop the model
        for i_t in range(self.nt):
            if i_t % self.n_out == 0:
                states_list.append(state)
            state, closure_state = step(
                self.dt, self.case, self.closure, state, closure_state,
                closure_parameters, swr_frac
            )
        time = jnp.arange(0, self.nt*self.dt, self.n_out*self.dt)

        # generate trajectory
        u_list = [s.u for s in states_list]
        v_list = [s.v for s in states_list]
        t_list = [s.t for s in states_list]
        s_list = [state.s for state in states_list]
        trajectory = Trajectory(
            self.init_state.grid, time, jnp.vstack(t_list), jnp.vstack(s_list),
            jnp.vstack(u_list), jnp.vstack(v_list)
        )
        return trajectory


@partial(jit, static_argnames=('dt', 'case', 'closure'))
def step(
        dt: float,
        case: Case,
        closure: Closure,
        state: State,
        closure_state: ClosureStateAbstract,
        closure_parameters: ClosureParametersAbstract,
        swr_frac: Float[Array, 'nz+1']
    ) -> Tuple[State, ClosureStateAbstract]:
    r"""
    Run one time-step of the model.
    

    This functions first call the closure to compute the eddy-diffusivity and
    viscosity, and then integrate the equations of tracers and momentum. It
    modifies the :code:`state` with these new values and then returns the new
    :code:`state` and :code:`closure_state`.

    Parameters
    ----------
    dt : float
        Time-step of integration for every iteration :math:`[\text s]`.
    case : Case
        Physical case and forcings of the experiment.
    closure : Closure
        Abstraction representing the chosen closure.
    state : State
        Curent state of the water column.
    closure_state : ClosureStateAbstract
        Curent state of the water column for the closure variables.
    closure_parameters : ClosureParametersAbstract
        A set of parameters of the used with the :code:`closure`.
    swr_frac : Float[~jax.Array, 'nz+1']
        Fraction of solar penetration throught the water column
        :math:`[\text{dimensionless}]`.

    Returns
    -------
    state : State
        State of the water column at next time-step.
    closure_state : ClosureStateAbstract
        State of the water column at next time-step for the closure variables.

    Note
    ----
    This function is jitted with JAX, it should make it faster, but the
    :func:`~jax.jit` decorator can be removed.
    """
    grid = state.grid

    # advance closure state (compute eddy-diffusivity and viscosity)
    closure_state = closure.step_fun(
        state, closure_state, dt, closure_parameters, case
    )

    # advance tracers
    t_new, s_new = advance_tra_ed(
        state.t, state.s, closure_state.akt, swr_frac, grid.hz, dt, case
    )

    # advance velocities
    u_new, v_new = advance_dyn_cor_ed(
        state.u, state.v, grid.hz, closure_state.akv, dt, case
    )

    # write the new state
    state = eqx.tree_at(lambda tree: tree.t, state, t_new)
    state = eqx.tree_at(lambda t: t.s, state, s_new)
    state = eqx.tree_at(lambda t: t.u, state, u_new)
    state = eqx.tree_at(lambda t: t.v, state, v_new)

    return state, closure_state


def lmd_swfrac(hz: Float[Array, 'nz']) -> Float[Array, 'nz+1']:
    r"""
    Compute solar forcing.

    Compute fraction of solar shortwave flux penetrating to specified depth due
    to exponential decay in Jerlov water type. This function is called once
    before running the model.

    Parameters
    ----------
    hz : Float[~jax.Array, 'nz']
        Thickness of cells from deepest to shallowest :math:`[\text m]`.

    Returns
    -------
    swr_frac : Float[~jax.Array, 'nz+1']
        Fraction of solar penetration throught the water column
        :math:`[\text{dimensionless}]`.
    """
    nz, = hz.shape
    mu1 = 0.35
    mu2 = 23.0
    r1 = 0.58
    attn1 = -1.0 / mu1
    attn2 = -1.0 / mu2

    xi1 = attn1 * hz
    xi2 = attn2 * hz

    def lax_step(sdwk, k):
        sdwk1, sdwk2 = sdwk
        sdwk1 = lax.cond(xi1[nz-k] > -20, lambda x: x*jnp.exp(xi1[nz-k]),
                             lambda x: 0.*x, sdwk1)
        sdwk2 = lax.cond(xi2[nz-k] > -20, lambda x: x*jnp.exp(xi2[nz-k]),
                             lambda x: 0.*x, sdwk2)
        return (sdwk1, sdwk2), sdwk1+sdwk2

    _, swr_frac = lax.scan(lax_step, (r1, 1.0 - r1), jnp.arange(1, nz+1))
    return jnp.concat((swr_frac[::-1], jnp.array([1.])))


def advance_tra_ed(
        t: Float[Array, 'nz'],
        s: Float[Array, 'nz'],
        akt: Float[Array, 'nz+1'],
        swr_frac: Float[Array, 'nz+1'],
        hz: Float[Array, 'nz'],
        dt: float,
        case: Case
    ) -> Tuple[Float[Array, 'nz'], Float[Array, 'nz']]:
    r"""
    Integrate vertical diffusion term for tracers.

    First the flux divergences are computed taking in account the forcings.
    Then the diffusion equation of the tracers system is solved, and the
    tracers at next time-step are returned. The solved equation is for
    :math:`C` a tracer :

    :math:`\partial _z ( K_m \partial _z C) + \partial _t C + F = 0`

    where :math:`F` is the representation of the forcings.

    Parameters
    ----------
    t : Float[~jax.Array, 'nz']
        Current temperature on the center of the cells :math:`[° \text C]`.
    s : Float[~jax.Array, 'nz']
        Current salinity on the center of the cells :math:`[\text{psu}]`.
    akt : Float[~jax.Array, 'nz+1']
        Current eddy-diffusivity :math:`K_m` on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    swr_frac : Float[~jax.Array, 'nz+1']
        Fraction of solar penetration throught the water column
        :math:`[\text{dimensionless}]`.
    hz : Float[~jax.Array, 'nz']
        Thickness of cells from deepest to shallowest :math:`[\text m]`.
    dt : float
        Time-step of the integration step :math:`[\text s]`.
    case : Case
        Physical case and forcings of the experiment.

    Returns
    -------
    t : Float[Array, 'nz']
        Temperature on the center of the cells at next step
        :math:`[° \text C]`.
    s : Float[Array, 'nz']
        Salinity on the center of the cells at next step :math:`[\text{psu}]`.
    """
    # 1 - Fluxes
    # Temperature
    fc_t = case.rflx_sfc_max * swr_frac
    fc_t = fc_t.at[-1].add(case.tflx_sfc)
    fc_t = fc_t.at[0].set(0.)
    # apply flux divergence
    dft = hz*t + dt*(fc_t[1:] - fc_t[:-1])
    # Salinity
    fc_s = jnp.zeros(t.shape[0]+1)
    fc_s.at[-1].set(case.sflx_sfc)
    # apply flux divergence
    dfs = hz*s + dt*(fc_s[1:] - fc_s[:-1])

    # 2 - Implicit integration for vertical diffusion
    # Temperature
    dft = dft.at[0].add(-dt * case.tflx_btm)
    t = diffusion_solver(akt, hz, dft, dt)
    # Salinity
    dfs = dfs.at[0].add(-dt * case.sflx_btm)
    s = diffusion_solver(akt, hz, dfs, dt)

    return t, s


def advance_dyn_cor_ed(
        u: Float[Array, 'nz'],
        v: Float[Array, 'nz'],
        hz: Float[Array, 'nz'],
        akv: Float[Array, 'nz+1'],
        dt: float,
        case: Case
    ) -> Tuple[Float[Array, 'nz'], Float[Array, 'nz']]:
    r"""
    Integrate vertical diffusion and Coriolis terms for momentum.

    First the Coriolis term is computed, then the momentum forcings are applied
    and finally, the diffusion equation is solved. The momentum at next time-
    step is returned. The equation which is solved is :

    :math:`\partial_z (K_v \partial_z U) + F_{\text{cor}}(U) + F = 0`

    where :math:`F_{\text{cor}}` represent the Coriolis effect, and :math:`F`
    represent the effect of the forcings.

    Parameters
    ----------
    u : Float[~jax.Array, 'nz']
        Current zonal velocity on the center of the cells
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    v : Float[~jax.Array, 'nz']
        Current meridional velocity on the center of the cells
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    akv : Float[~jax.Array, 'nz+1']
        Current eddy-viscosity :math:`K_v` on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    hz : Float[~jax.Array, 'nz']
        Thickness of cells from deepest to shallowest :math:`[\text m]`.
    dt : float
        Time-step of the integration step :math:`[\text s]`.
    case : Case
        Physical case and forcings of the experiment.

    Returns
    -------
    u : Float[Array, 'nz']
        Zonal velocity on the center of the cells at the next time step
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    v : Float[Array, 'nz']
        Meridional velocity on the center of the cells at the next time step
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    """
    gamma_cor = 0.55
    fcor = case.fcor

    # 1 - Compute Coriolis term
    cff = (dt * fcor) ** 2
    cff1 = 1 / (1 + gamma_cor * gamma_cor * cff)
    fu = cff1 * hz * ((1-gamma_cor*(1-gamma_cor)*cff)*u + dt*fcor*v)
    fv = cff1 * hz * ((1-gamma_cor*(1-gamma_cor)*cff)*v - dt*fcor*u)

    # 2 - Apply surface and bottom forcing
    fu = fu.at[-1].add(dt * case.ustr_sfc)
    fv = fv.at[-1].add(dt * case.vstr_sfc)
    fu = fu.at[0].add(-dt * case.ustr_btm)
    fv = fv.at[0].add(-dt * case.vstr_btm)

    # 3 - Implicit integration for vertical viscosity
    u = diffusion_solver(akv, hz, fu, dt)
    v = diffusion_solver(akv, hz, fv, dt)

    return u, v


def diffusion_solver(
        ak: Float[Array, 'nz+1'],
        hz: Float[Array, 'nz'],
        f: Float[Array, 'nz'],
        dt: float
    ) -> Float[Array, 'nz']:
    r"""
    Solve a diffusion problem with finite volumes.

    The diffusion problems can be written

    :math:`\partial _z (K \partial _z X) + \dfrac f {\Delta t \Delta x} = 0`

    where we are searching for :math:`X` and where :math:`f` represents the
    temporal derivative and forcings. This function transforms this problem in
    a tridiagonal system and then solve it.

    Parameters
    ----------
    ak : Float[~jax.Array, 'nz+1']
        Diffusion at the cell interfaces :math:`K` in
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    hz : Float[~jax.Array, 'nz']
        Thickness of cells from deepest to shallowest
        :math:`\left[\text m\right]`.
    f : Float[~jax.Array, 'nz']
        Right-hand flux of the equation :math:`f` in
        :math:`[[X] \cdot \text m ]`.
    dt : float
        Time-step of discretisation :math:`[\text s]`.
    
    Returns
    -------
    x : Float[~jax.Array, 'nz']
        Solution of the diffusion problem
        :math:`X` in :math:`\left[[X]\right]`.
    """
    # fill the coefficients for the tridiagonal matrix
    a_in = -2.0 * dt * ak[1:-2] / (hz[:-2] + hz[1:-1])
    c_in = -2.0 * dt * ak[2:-1] / (hz[2:] + hz[1:-1])
    b_in = hz[1:-1] - a_in - c_in

    # bottom boundary condition
    c_btm = -2.0 * dt * ak[1] / (hz[1] + hz[0])
    b_btm = hz[0] - c_btm

    # surface boundary condition
    a_sfc = -2.0 * dt * ak[-2] / (hz[-2] + hz[-1])
    b_sfc = hz[-1] - a_sfc

    # concatenations
    a = add_boundaries(0., a_in, a_sfc)
    b = add_boundaries(b_btm, b_in, b_sfc)
    c = add_boundaries(c_btm, c_in, 0.)

    x = tridiag_solve(a, b, c, f)

    return x



"""
Physical parameters and forcings.

This module comes down to Case class. This class can be obtained by the prefix
:code:`tunax.case.` or directly by :code:`tunax.`.

"""

from __future__ import annotations

import equinox as eqx
import jax.numpy as jnp
from jax import device_get


_OMEGA = 7.292116e-05
"""float: Rotation rate of the Earth [rad.s-1]."""
_RAD_DEG = jnp.pi/180.
"""float: Measure of one degree in radiant [rad.°-1]."""


class Case(eqx.Module):
    r"""
    Physical parameters and forcings.

    This class contains all the physical constants, and the constant forcings
    that definine an experience for the model.

    Parameters
    ----------
    rho0 : float, default=1024.
        cf. attribute.
    grav : float, default=9.81
        cf. attribute.
    cp : float, default=3985.
        cf. attribute.
    alpha : float, default=2e-4
        cf. attribute.
    beta : float, default=8e-4
        cf. attribute.
    t_rho_ref : float, default=0.
        cf. attribute.
    s_rho_ref : float, default=35.
        cf. attribute.
    vkarmn : float, default=0.384
        cf. attribute.
    fcor : float, default=0.
        cf. attribute.
    ustr_sfc : float, default=0.
        cf. attribute.
    ustr_btm : float, default=0.
        cf. attribute.
    vstr_sfc : float, default=0.
        cf. attribute.
    vstr_btm : float, default=0.
        cf. attribute.
    tflx_sfc : float, default=0.
        cf. attribute.
    tflx_btm : float, default=0.
        cf. attribute.
    sflx_sfc : float, default=0.
        cf. attribute.
    sflx_btm : float, default=0.
        cf. attribute.
    rflx_sfc_max : float, default=0.
        cf. attribute.

    Attributes
    ----------
    rho0 : float, default=1024.
        Default density of saltwater :math:`[\text{kg} \cdot \text{m}^{-3}]`.
    grav : float, default=9.81
        Gravity acceleration :math:`[\text{m} \cdot \text{s}^{-2}]`.
    cp : float, default=3985.
        Specific heat capacity of saltwater
        :math:`[\text{J} \cdot \text{kg}^{-1} \cdot \text{K}^{-1}]`.
    alpha : float, default=2e-4
        Thermal expansion coefficient :math:`[\text{K}^{-1}]`.
    beta : float, default=8e-4
        Salinity expansion coefficient :math:`[\text{psu}^{-1}]`.
    t_rho_ref : float, default=0.
        Reference temperature for the density computation :math:`[° \text C]`.
    s_rho_ref : float, default=35.
        Reference salinity for the density computation :math:`[\text{psu}]`.
    vkarmn : float, default=0.384
        Von Kármán constant :math:`[\text{dimensionless}]`.
    fcor : float, default=0.
        Coriolis frequency at the water column
        :math:`[\text{rad} \cdot \text{s}^{-1}]`.
    ustr_sfc : float, default=0.
        Zonal wind stress :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    ustr_btm : float, default=0.
        Zonal current stress at the bottom
        :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    vstr_sfc : float, default=0.
        Meridional wind stress :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    vstr_btm : float, default=0.
        Meridional current stress at the bottom
        :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    tflx_sfc : float, default=0.
        Non-penetrative heat flux at the surface
        :math:`[\text{K} \cdot \text{m} \cdot \text{s}^{-1}]`.
    tflx_btm : float, default=0.
        Non-penetrative heat flux at the bottom
        :math:`[\text{K} \cdot \text{m} \cdot \text{s}^{-1}]`.
    sflx_sfc : float, default=0.
        Fresh water flux at the surface
        :math:`[\text{psu} \cdot \text{m} \cdot \text{s}^{-1}]`.
    sflx_btm : float, default=0.
        Fresh water flux at the bottom
        :math:`[\text{psu} \cdot \text{m} \cdot \text{s}^{-1}]`.
    rflx_sfc_max : float, default=0.
        Maximum solar radiation flux at the surface
        :math:`[\text{K} \cdot \text{m} \cdot \text{s}^{-1}]`.

    """

    # physcal constants
    rho0: float = 1024.
    grav: float = 9.81
    cp: float = 3985.
    alpha: float = 2e-4
    beta: float = 8e-4
    t_rho_ref: float = 0.
    s_rho_ref: float = 35.
    vkarmn: float = 0.384
    #forcings
    fcor: float = 0.
    ustr_sfc: float = 0.
    ustr_btm: float = 0.
    vstr_sfc: float = 0.
    vstr_btm: float = 0.
    tflx_sfc: float = 0.
    tflx_btm: float = 0.
    sflx_sfc: float = 0.
    sflx_btm: float = 0.
    rflx_sfc_max: float = 0.

    def set_lat(self, lat: float) -> Case:
        """
        Set the Coriolis frequency from the latitude.

        Parameters
        ----------
        lat : float
            Latitude of the water column :math:`[°]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the the new value of :attr:`fcor`.
        """
        fcor = float(device_get(2.*_OMEGA*jnp.sin(_RAD_DEG*lat)))
        print(type(fcor))
        print(fcor)
        case = eqx.tree_at(lambda t: t.fcor, self, fcor)
        return case

    def set_u_wind(self, u_wind: float) -> Case:
        r"""
        Set the zonal wind stress with the zonal wind speed.

        Parameters
        ----------
        u_wind : float
            Zonal wind speed :math:`[\text m \cdot \text s ^{-1}]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the new value of :attr:`ustr_sfc`.
        """
        case = eqx.tree_at(lambda t: t.ustr_sfc, self, u_wind**2)
        return case

    def set_u_cur(self, u_cur: float) -> Case:
        r"""
        Set the zonal current stress with the zonal current.

        Parameters
        ----------
        u_cur : float
            Zonal current speed :math:`[\text m \cdot \text s ^{-1}]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the new value of :attr:`ustr_btm`.
        """
        case = eqx.tree_at(lambda t: t.ustr_btm, self, u_cur**2)
        return case

    def set_v_wind(self, v_wind: float) -> Case:
        r"""
        Set the meridional wind stress with the meridional wind.

        Parameters
        ----------
        v_wind : float
            Meridional wind speed :math:`[\text m \cdot \text s ^{-1}]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the new value of :attr:`vstr_sfc`.
        """
        case = eqx.tree_at(lambda t: t.vstr_sfc, self, v_wind**2)
        return case

    def set_v_cur(self, v_cur: float) -> Case:
        r"""
        Set the meridional current stress with the meridional current.

        Parameters
        ----------
        v_cur : float
            Meridional current speed :math:`[\text m \cdot \text s ^{-1}]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the new value of :attr:`vstr_btm`.
        """
        case = eqx.tree_at(lambda t: t.vstr_btm, self, v_cur**2)
        return case

    def set_tpw_sfc(self, tpw_sfc: float) -> Case:
        r"""
        Set the heat flux at surface from the heat power.

        Parameters
        ----------
        tpw_sfc : float
            Non-penetrative heat power at the surface
            :math:`[\text W \cdot \text m ^{-2}]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the new value of :attr:`tflx_sfc`.
        """
        tflx_sfc = tpw_sfc/(self.rho0*self.cp)
        case = eqx.tree_at(lambda t: t.tflx_sfc, self, tflx_sfc)
        return case

    def set_tpw_btm(self, tpw_btm: float) -> Case:
        r"""
        Set the heat flux at bottom from the heat power.

        Parameters
        ----------
        tpw_btm : float
            Non-penetrative heat power at the bottom
            :math:`[\text W \cdot \text m ^{-2}]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the new value of :attr:`tflx_btm`.
        """
        tflx_btm = tpw_btm/(self.rho0*self.cp)
        case = eqx.tree_at(lambda t: t.tflx_btm, self, tflx_btm)
        return case

    def set_rpw_sfc_max(self, rpw_sfc_max: float) -> Case:
        r"""
        Set the maximum solar radiation from the solar power

        Parameters
        ----------
        rpw_sfc_max : float
            Maximum solar radiation power at the surface (penetrative)
            :math:`[\text W \cdot \text m ^{-2}]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the new value of :attr:`rflx_sfc_max`.
        """
        rflx_sfc_max = rpw_sfc_max/(self.rho0*self.cp)
        case = eqx.tree_at(lambda t: t.rflx_sfc_max, self, rflx_sfc_max)
        return case



"""
Geometry and variables of the model.

This module contains the objects that are used in Tunax to describe the
geometry of the water column in :class:`Grid`, the variables of the water
column at one time-step in :class:`State` and the time-series of the model
computation in :class:`Trajectories`. These classes can be obtained by the
prefix :code:`tunax.space.` or directly by :code:`tunax.`.

"""

from __future__ import annotations

import equinox as eqx
import xarray as xr
import jax.numpy as jnp
from jax import vmap
from jaxtyping import Float, Array


def _piecewise_linear_ramp(z: float, z0: float, f0: float)-> float:
    r"""
    Mathemacial function used for the state initialisation.

    Apply to z a function linear by part and continuous :
    f(z) = 0 if z < zm and f(z) = f0 (1-z/zm) else.

    Parameters
    ----------
    z : float
        The value where to apply the function.
    zm : float
        The point of connexion of the two linear parts of the function.
    f0 : float
        The value of the function in 0.

    Returns
    -------
    fz : float
        The value of the function in z.
    """
    return f0*(z/-z0+1) * (z>z0)

def _piecewise_linear_flat(z: float, zm: float, f0: float, sl: float) -> float:
    r"""
    Mathemacial function used for the state initialisation.

    Apply to z a function linear by part and continuous :
    f(z) = f0 + s_l (z-zm) if z < z_m and f(z) = f0 else.

    Parameters
    ----------
    z : float
        The value where to apply the function.
    zm : float
        The point of connexion of the two linear parts of the function.
    f0 : float
        The value of the function in 0 and in the right part of the
        funcion.
    sl : float
        The slope of the left part of the function.

    Returns
    -------
    fz : float
        The value of the function in z.
    """
    return f0 + sl*(z-zm) * (z<zm)


class Grid(eqx.Module):
    r"""
    One dimensional spatial geometry of a water column.

    This mesh is made up of a number of :attr:`nz` of cells (:attr:`zr`) of
    potentially varying thickness (:attr:`hz`), separated by interface points
    (:attr:`zw`) and extending from the ocean surface at a depth of :math:`0`
    to the ocean floor at a depth of :attr:`hbot`.

    Parameters
    ----------
    zr : Float[~jax.Array, 'nz']
        cf. attribute.
    zw : Float[~jax.Array, 'nz+1']
        cf. attribute.

    Attributes
    ----------
    nz : int
        Number of cells.
    hbot : float
        Depth of the water column :math:`[\text m]`.
    zr : Float[~jax.Array, 'nz']
        Depths of cell centers from deepest to shallowest :math:`[\text m]`
    zw : Float[~jax.Array, 'nz+1']
        Depths of cell interfaces from deepest to shallowest :math:`[\text m]`.
    hz : Float[~jax.Array, 'nz']
        Thickness of cells from deepest to shallowest :math:`[\text m]`.

    Note
    ----
    The constructor :code:`__init__` takes only :attr:`zr` and :attr:`zw` as
    as arguments and construct the other attributes from them. The centers of
    the cells :attr:`zr` are not necessarly the middle between the interfaces
    :attr:`zw`.

    """

    nz: int
    hbot: float
    zr: Float[Array, 'nz']
    zw: Float[Array, 'nz+1']
    hz: Float[Array, 'nz']

    def __init__(self, zr: Float[Array, 'nz'], zw: Float[Array, 'nz+1']):
        self.nz = zr.shape[0]
        self.hbot = zw[0]
        self.zw = zw
        self.zr = zr
        self.hz = zw[1:] - zw[:-1]

    def find_index(self, h: float) -> int:
        r"""
        Find the index of a depth.

        Find the index :code:`i` so that the depth :code:`h` is in cell
        :code:`i`, which means :math:`z^w_i \leqslant -h \leqslant z^w_{i+1}`
        if :math:`h \leqslant 0` and :math:`i=-1` if :math:`h>0`.

        Parameters
        ----------
        h : float, positive
            The depth to search the index :math:`[\text m]`.

        Returns
        -------
        i : int
            The index corresponding to the depth :code:`h`.
        """
        return int(jnp.searchsorted(self.zw, -h, side='right')) - 1

    @classmethod
    def linear(cls, nz: int, hbot: int) -> Grid:
        r"""
        Creates a grid with equal thickness cells.

        The grid instance will have :attr:`nz` cells of equal thickness for a
        depth of :attr:`hbot`.

        Parameters
        ----------
        nz : int
            Number of cells.
        hbot : float, positive
            Depth of the water column :math:`[\text m]`.
        """
        zw = jnp.linspace(-hbot, 0, nz+1)
        zr = 0.5*(zw[:-1]+zw[1:])
        return cls(zr, zw)

    @classmethod
    def analytic(
            cls,
            nz: int,
            hbot: float,
            hc: float,
            theta: float=6.5
        ) -> Grid:
        r"""
        Creates a grid of type analytic.

        The grid instance will have a depth of :attr:`hbot` and :attr:`nz`
        cells of thickness almost equals above :code:`hc` and wider under, the
        strecht parameter being defined by :code:`theta`.

        Parameters
        ----------
        nz : int
            Number of cells.
        hbot : float, positive
            Depth of the water column :math:`[\text m]`.
        hc : float, positive
            Reference depth :math:`[\text m]`.
        theta : float, default=6.5
            Stretching parameter toward the surface
            :math:`[\text{dimensionless}]`.
        """
        sc_w = jnp.linspace(-1, 0, nz+1)
        sc_r = (sc_w[:-1] + sc_w[1:])/2
        cs_r = (1-jnp.cosh(theta*sc_r))/(jnp.cosh(theta)-1)
        cs_w = (1-jnp.cosh(theta*sc_w))/(jnp.cosh(theta)-1)
        zw = (hc*sc_w + hbot*cs_w) * hbot/(hbot+hc)
        zr = (hc*sc_r + hbot*cs_r) * hbot/(hbot+hc)
        return cls(zr, zw)

    @classmethod
    def orca75(cls, hbot: float) -> Grid:
        r"""
        Creates the ORCA 75 grid from NEMO.

        The whole grid is created then levels between the depth :attr:`hbot`
        and :math:`0` are extracted.

        Parameters
        ----------
        hbot : float, positive
            Depth of the water column :math:`[\text m]`.
        """
        nz_orca = 75
        zsur = -3958.95137127683
        za2 = 100.7609285
        za0 = 103.9530096
        za1 = 2.415951269
        zkth = 15.3510137
        zkth2 = 48.02989372
        zacr = 7.
        zacr2 = 13.
        sc_r = jnp.arange(nz_orca-0.5, 0.5, -1)
        sc_w = jnp.arange(nz_orca, 0, -1)
        zw_orca = -(zsur + za0*sc_w + za1*zacr*jnp.log(jnp.cosh((sc_w-zkth)/\
            zacr)) + za2*zacr2*jnp.log(jnp.cosh((sc_w-zkth2)/zacr2)))
        zr_orca = -(zsur + za0*sc_r + za1*zacr*jnp.log(jnp.cosh((sc_r-zkth)/\
            zacr)) + za2*zacr2*jnp.log(jnp.cosh((sc_r-zkth2)/zacr2)))
        ibot = jnp.argmin(zw_orca <= -hbot)
        if ibot == 0:
            ibot = 1
        zw_orca = zw_orca.at[-1].set(0.)
        zw = zw_orca[ibot-1:]
        zr = zr_orca[ibot-1:]
        return cls(zr, zw)

    @classmethod
    def load(cls, ds: xr.Dataset) -> Grid:
        """
        Creates the grid defined by a dataset :code:`ds` of an observation.

        The dataset must be formated to have the variables corresponding to
        :attr:`zr` and :attr:`zw`.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset from which to extract the grid.
        """
        zw = jnp.array(ds['zw'], dtype=jnp.float32)
        zr = jnp.array(ds['zr'], dtype=jnp.float32)
        return cls(zr, zw)


class State(eqx.Module):
    r"""
    Water column state at one time-step.

    This state is defined on a :attr:`grid` describing the geometry, and is
    composed of the variables of the water column : the values of the momentum
    and the tracers on this :attr:`grid`. The call of the constructor build a
    state on the :attr:`grid` with all the variables set to :math:`0`.

    Parameters
    ----------
    grid : Grid
        cf. attribute.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column.
    u : Float[~jax.Array, 'nz']
        Zonal velocity on the center of the cells :math:`\left[\text m \cdot
        \text s^{-1}\right]`.
    v : Float[~jax.Array, 'nz']
        Meridional velocity on the center of the cells :math:`\left[\text m
        \cdot \text s^{-1}\right]`.
    t : Float[~jax.Array, 'nz']
        Temperature on the center of the cells :math:`[° \text C]`.
    s : Float[~jax.Array, 'nz']
        Salinity on the center of the cells :math:`[\text{psu}]`.

    """

    grid: Grid
    u: Float[Array, 'nz']
    v: Float[Array, 'nz']
    t: Float[Array, 'nz']
    s: Float[Array, 'nz']

    @classmethod
    def zeros(cls, grid: Grid) -> State:
        """
        Initialize an instance with all variables equals to zero from a grid.

        Parameters
        ----------
        grid : Grid
            Geometry of the water column.

        Returns
        -------
        state : State
            An instance defined on the grid with all variables set to 0.        
        """
        u = jnp.zeros(grid.nz)
        v = jnp.zeros(grid.nz)
        t = jnp.zeros(grid.nz)
        s = jnp.zeros(grid.nz)
        return State(grid, u, v, t, s)

    def init_u(self, hmxl: float=20., u_sfc: float=0.) -> State:
        r"""
        Initialize zonal velocity with a classical wind stratification.

        Return a State object where :attr:`u` is continuous and linear by part
        :math:`u(z) = \begin{cases}
        0 & \text{if } z < h_{\text{mxl}}\\
        u_{\text{sfc}} \left( 1 - \dfrac z {h_{\text{mxl}}}\right) &
        \text{else} \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        u_sfc : float, default=0.
            Surface zonal velocity :math:`\left[\text m \cdot \text
            s^{-1}\right]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of zonal velocity.
        """
        maped_fun = vmap(_piecewise_linear_ramp, in_axes=(0, None, None))
        u_new = maped_fun(self.grid.zr, -hmxl, u_sfc)
        return eqx.tree_at(lambda t: t.u, self, u_new)

    def init_v(self, hmxl: float=20., v_sfc: float=0.) -> State:
        r"""
        Initialize meridional velocity with a classical wind stratification.

        Return a State object where :attr:`v` is continuous and linear by part
        :math:`v(z) = \begin{cases}
        0 & \text{if } z < h_{\text{mxl}}\\
        v_{\text{sfc}} \left( 1 - \dfrac z {h_{\text{mxl}}}\right) &
        \text{else} \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        u_sfc : float, default=0.
            Surface meridional velocity :math:`\left[\text m \cdot \text
            s^{-1}\right]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of meridional
            velocity.
        """
        maped_fun = vmap(_piecewise_linear_ramp, in_axes=(0, None, None))
        v_new = maped_fun(self.grid.zr, -hmxl, v_sfc)
        return eqx.tree_at(lambda t: t.v, self, v_new)

    def init_t(
            self,
            hmxl: float=20.,
            t_sfc: float=21.,
            strat_t: float=5.1e-2
        ) -> State:
        r"""
        Initialize temperature with a classical tracer stratification.

        Return a State object where :attr:`t` is linear by part and continous
        :math:`T(z) = \begin{cases}
        t_{\text{sfc}} + S_T(z-h_{\text{mxl}}) & \text{if } z <
        h_{\text{mxl}}\\
        t_{\text{sfc}} & \text{else}
        \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        t_sfc : float, default=21.
            Surface temperature :math:`[° \text C]`.
        strat_t : float, default=5.1e-2
            Thermal stratification above the mixed layer noted by :math:`S_T`
            :math:`[\text K \cdot \text m ^{-1}]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of temperature.
        """
        maped_fun = vmap(_piecewise_linear_flat, in_axes=(0, None, None, None))
        t_new = maped_fun(self.grid.zr, -hmxl, t_sfc, strat_t)
        return eqx.tree_at(lambda tree: tree.t, self, t_new)

    def init_s(
            self,
            hmxl: float=20.,
            s_sfc: float=35.,
            strat_s: float=1.3e-2
        ) -> State:
        r"""
        Initialize salinity with a classical tracer stratification.

        Return a State object where :attr:`s` is linear by part and continous
        :math:`S(z) = \begin{cases}
        s_{\text{sfc}} + S_S(z-h_{\text{mxl}}) & \text{if } z <
        h_{\text{mxl}}\\
        s_{\text{sfc}} & \text{else}
        \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        s_sfc : float, default=21.
            Surface salinity :math:`[\text{psu}]`.
        strat_t : float, default=5.1e-2
            Salinity stratification above the mixed layer noted by :math:`S_T`
            :math:`[\text{psu} \cdot \text m ^{-1}]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of temperature.
        """
        maped_fun = vmap(_piecewise_linear_flat, in_axes=(0, None, None, None))
        s_new = maped_fun(self.grid.zr, -hmxl, s_sfc, strat_s)
        return eqx.tree_at(lambda t: t.s, self, s_new)


class Trajectory(eqx.Module):
    r"""
    Define the history of a simulation or an observation.

    Contains the timeseries of the momentum and the tracers throught the space
    of the :attr:`grid` and the :attr:`time`.

    Parameters
    ----------
    grid : Grid
        cf. attribute.
    time : Float[~jax.Array, 'nt']
        cf. attribute.
    u : Float[~jax.Array, 'nt nz']
        cf. attribute.
    v : Float[~jax.Array, 'nt nz']
        cf. attribute.
    t : Float[~jax.Array, 'nt nz']
        cf. attribute.
    s : Float[~jax.Array, 'nt nz']
        cf. attribute.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column.
    time : Float[~jax.Array, 'nt']
        Time at each steps of observation from the begining of the simulation
        :math:`[\text s]`.
    u : Float[~jax.Array, 'nt nz']
        Time-serie of zonal velocity :math:`\left[\text m \cdot \text
        s^{-1}\right]`.
    v : Float[~jax.Array, 'nt nz']
        Time-serie of meridional velocity :math:`\left[\text m \cdot \text
        s^{-1}\right]`.
    t : Float[~jax.Array, 'nt nz']
        Time-serie of temperature :math:`[\text C°]`.
    s : Float[~jax.Array, 'nt nz']
        Time-serie of salinity :math:`[\text{psu}]`.

    """

    grid: Grid
    time: Float[Array, 'nt']
    t: Float[Array, 'nt nz']
    s: Float[Array, 'nt nz']
    u: Float[Array, 'nt nz']
    v: Float[Array, 'nt nz']

    def to_ds(self) -> xr.Dataset:
        """
        Exports the trajectory in an xarray.Dataset.

        The dimensions of the dataset are :attr:`time`, :code:`grid.zr` and
        :code:`grid.zw`, the variables are :attr:`u`, :attr:`v`, :attr:`t` and
        :attr:`s`, all defined on the dimensions (:attr:`time`, :code:`zr`).

        Returns
        -------
        ds : xarray.Dataset
            Dataset of the trajectory.
        """
        variables = {'u': (('time', 'zr'), self.u),
                     'v': (('time', 'zr'), self.v),
                     't': (('time', 'zr'), self.t),
                     's': (('time', 'zr'), self.s)}
        coords = {'time': self.time,
                  'zr': self.grid.zr,
                  'zw': self.grid.zw}
        return xr.Dataset(variables, coords)

    def extract_state(self, i_time: int) -> State:
        """
        Extracts the water column state at one time index.

        Parameters
        ----------
        i_time : int
            The time index of the moment to extract the state.

        Returns
        -------
        state : State
            The state of the trajectory at the time of index :code:`i_time`.
        """
        return State(
            self.grid, self.u[i_time, :], self.v[i_time, :], self.t[i_time, :],
            self.s[i_time, :],
        )



"""
Abstraction for the calibration kernel.

This module contain the main class :class:`Fitter` which is the code of Tunax
for the calibration of the closure parts. The classes
:class:`FittableParameter` and :class:`FittableParametersSet` are used to make
a link between the optimization part and the closures structures. These class
can be obtained by the prefix :code:`tunax.fitter.` or directly by
:code:`tunax.`.

"""

from __future__ import annotations
from typing import Callable, Optional, List, Dict

import optax
import equinox as eqx
import numpy as np
import jax.numpy as jnp
from jaxtyping import Float, Array
from jax import grad

from tunax.space import Trajectory
from tunax.closure import ClosureParametersAbstract, Closure
from tunax.closures_registry import CLOSURES_REGISTRY
from tunax.model import SingleColumnModel
from tunax.database import Database


class FittableParameter(eqx.Module):
    """
    Calibration configuration for one parameter.

    An instance of this class must be created for every parameter of the
    closure that will be calibrated and for every not default values during
    calibration.

    Parameters
    ----------
    do_fit : bool
        cf. attribute.
    val : float, default=0.
        cf. attribute.

    Attributes
    ----------
    do_fit : bool
        The parameter will be calibrated.
    val : float, default=0.
        If :attr:`do_fit` : the initial value for calibration (at the first
        step of the calibration) ; if :attr:`do_fit` is false : the constant
        value to take for this parameter if it's not the default one in the
        closure.

    """

    do_fit: bool
    val: float = 0.


class FittableParametersSet(eqx.Module):
    """
    Complete closure calibration parameters.

    This class is the set of all the configurations on the closure parameters
    for the calibration. It makes the link between the array on which the
    optimizer works and the closure parameters class.

    Parameters
    ----------
    coef_fit_dict : Dict[str, FittableParameter]
        cf. attribute.
    closure_name : str
        Name of the chosen closure, must be a key of
        :attr:`~closures_registry.CLOSURES_REGISTRY`, see its documentation
        for the available closures.

    Attributes
    ----------
    coef_fit_dico : Dict[str, FittableParameter]
        The set of all the configurations of all the parameters that will be
        calibrated and the one constants but not with the default value of the
        closure.
    closure : Closure
        The abstraction that represent the used closure.
    
    """

    coef_fit_dict: Dict[str, FittableParameter]
    closure: Closure

    def __init__(
            self,
            coef_fit_dict: Dict[str, FittableParameter],
            closure_name: str
        ) -> FittableParametersSet:
        self.coef_fit_dict = coef_fit_dict
        self.closure = CLOSURES_REGISTRY[closure_name]

    @property
    def n_calib(self) -> int:
        """
        Number of variables that are calibrated.

        Returns
        -------
        nc : int
            Number of variables that are calibrated.
        """
        nc = 0
        for coef_fit in self.coef_fit_dict.values():
            if coef_fit.do_fit:
                nc += 1
        return nc

    def fit_to_closure(
            self,
            x: Float[Array, 'nc']
        ) -> ClosureParametersAbstract:
        """
        Transforms an fitted array in a set of closure parameters.

        This method copy the fixed non-default values of :attr:`coef_fit_dict`
        and copy the calibrated values from :code:`x`. Which is simply the
        parameters values in the order that is indicated by
        :attr:`coef_fit_dict`.

        Parameters
        ----------
        x : Float[~jax.Array, 'nc']
            The array on which the optimize works to find the best values. It
            is the array of the parameters that are calibrated.

        Returns
        -------
        clo_params : ClosureParametersAbstract
            The instance of the closure parameters class (child class of
            :class:`~closure.ClosureParametersAbstract`) with the modifications
            of the calibration step.
        """
        clo_coef_dico = {}
        i_x = 0
        for coef_name, coef_fit in self.coef_fit_dict.items():
            if coef_fit.do_fit:
                clo_coef_dico[coef_name] = x[i_x]
                i_x += 1
            else:
                clo_coef_dico[coef_name] = coef_fit.val
        return self.closure.parameters_class(**clo_coef_dico)

    def gen_init_val(self) -> Float[Array, 'nc']:
        """
        Produce the fitted array for the first calibration step.

        This method simply copy the initial values of the calibrated
        coefficients in an array :code:`x` which will be used as the first
        calibration step for the optimizer.

        Returns
        -------
        x : Float[~jax.Array, 'nc']
            The initial vector for the optimizer at the first step of
            calibration.
        """
        x = []
        for coef_fit in self.coef_fit_dict.values():
            if coef_fit.do_fit:
                x.append(coef_fit.val)
        return jnp.array(x)


class Fitter(eqx.Module):
    r"""
    Reprensentation of a complete calibration configuration.

    A fitter is a link between the calibrated parameters configuration
    :attr:`coef_fit_params`, a :attr:`database` of observations to fit the
    model on, a loss function :attr:`loss` and some optimizer parameters. An
    instance can be call (with no parameters) to run the calibration. The
    :code:`__init__` method build a set of models corresponding to the given
    time-step :code:`dt` , the given closure :code:`closure_name` and the
    different initial states and physical cases extracted from the
    :attr:`database`.

    Parameters
    ----------
    coef_fit_params : FittableParametersSet
        cf. attribute.
    database : Database
        cf. attribute.
    dt : float
        The time-step used for defininf the forward model that will be
        calibrated :math:`[\text s]`.
    loss : Callable[[List[Trajectory], Database], float]
        cf. attribute.
    nloop : int, default=100
        cf. attribute.
    nit_loss : int, default = -1
        cf. attribute.
    learning_rate : float, default=0.001
        cf. attribute.
    verbatim : bool, default = True
        cf. attribute.
    output_path : Optional[str], default = '.'
        cf. attribute.

    Attributes
    ----------
    coef_fit_params: FittableParametersSet
        Parametrization of the closure parameters that must be calibrated and
        the one which are fixed with non-default values.
    database: Database
        Database of *observation* used for calibration, the optimizer will
        make the model fit to them.
    model_list : List[SingleColumnModel]
        List of model instances that represent the physical case and the
        initial condition for every *observation* in the database. At each
        calibration step, they will be call to compute the loss function.
    loss: Callable[[List[Trajectory], Database], float]
        Abstraction that the user create to describe its own loss function
        which will be minimized by the fitter. The function should represent
        how much the model with its closure fits to the :attr:`database`

        Parameters
        ----------
        trajectories : List[Trajectory]
            List of trajectories computed by the model and corresponding to
            each observation of the :attr:`database` case in the same order.
        database: Database
            cf.parameter
        
        Returns
        -------
        loss : float, positive
            The quantity that will be minimized by the fitter. The user must
            compute a quantity that compares the :code:`trajectories` done by
            the forward model (with the current values for the parameters of
            the closure at the current calibration step), and the trajectories
            from the :attr:`database`.

    nloop : int
        Maximum number of calibration loops.
    nit_loss : int, default = -1
        Number of iterations of every computation of the loss function in the
        output for diagnostic. Set to -1 to never compute it. Useless if
        :attr:`output_path` is :code:`None`.
    learning_rate : float, default=0.001
        Learning rate of the optimizer algorithm : how much it is fast at each
        step.
    verbatim : bool, default = True
        Print in the terminal the evolution of the calibration.
    output_path : Optional[str], default = '.'
        If :code:`None`, don't write the evolution on numpy files ; else must
        finishes by :code:`.npz` : in this file the compressed numpy output
        will be written.

    Note
    ----
    - During the calibration the gradient of the loss function is computed at
      every iterations, but not the loss function itself. The set of
      :attr:`nit_loss` increase the cost of the iteration if the value of the
      cost funtion is computed too often.
    
    - The output is written at every step so its readable by another python
      kernel during the calibration. To access to this data, one have to read
      the :code:`.npz` file with the :func:`numpy.load` function (with
      :code:`allow_pickle` set one :code:`True`), and then access to the
      evolutions of the calibrated vector (the closures paramters that are
      calibrated) with :code:`['x']`, to their gradients with :code:`['grads']`
      these are 2 dimensional arrays, the first dimension being the calibration
      iterations and the second one the parameters list. If :attr:`nit_loss` is
      not equal to -1, one can access to the evolution of the loss function
      with :code:`['loss_it']` which records the indexes of the iterations
      where the loss function is computed and :code:`['loss_values']` the
      corresponding values of the loss function.

    """

    coef_fit_params: FittableParametersSet
    database: Database
    model_list: List[SingleColumnModel]
    loss: Callable[[List[Trajectory], Database], float]
    nloop: int = 100
    nit_loss: int = -1
    learning_rate: float = 0.001
    verbatim: bool = True
    output_path: Optional[str] = eqx.field(default='.')

    def __init__(
            self,
            coef_fit_params: FittableParametersSet,
            database: Database,
            dt: float,
            loss: Callable[[List[Trajectory], Database], float],
            nloop: int = 100,
            nit_loss: int = -1,
            learning_rate: float = 0.001,
            verbatim: bool = True,
            output_path: Optional[str] = './'
        ) -> Fitter:
        # same attributes
        self.coef_fit_params = coef_fit_params
        self.database = database
        self.loss = loss
        self.nloop = nloop
        self.nit_loss = nit_loss
        self.learning_rate = learning_rate
        self.verbatim = verbatim
        self.output_path = output_path
        # building models list
        model_list = []
        for obs in self.database.observations:
            traj = obs.trajectory
            init_state = traj.extract_state(0)
            time = traj.time
            # extract time configuration from the trajectories of the database
            out_dt = float(time[1] - time[0])
            time_frame = float((time[-1] + out_dt)/3600.)
            closure_name = coef_fit_params.closure.name
            model = SingleColumnModel(
                time_frame, dt, out_dt, init_state, obs.case, closure_name
            )
            model_list.append(model)
        self.model_list = model_list
        # write the initialized values
        if output_path is not None:
            nc = coef_fit_params.n_calib
            empty_arr_2d = np.array([[] for _ in range(nc)])
            empty_arr_1d = np.array([])
            np.savez(
                output_path, x=empty_arr_2d, grads=empty_arr_2d,
                loss_it=empty_arr_1d, loss_values=empty_arr_1d
            )

    def loss_wrapped(self, x: Float[Array, 'nc']):
        """
        Wrapping of :attr:`loss` that takes only an array in argument.

        This method runs every model for each observations with the set of
        closure parameters corresponding to x, then it computes and returns the
        the value of the loss function.

        Parameters
        ----------
        x : Float[~jax.Array, 'nc']
            An array that represent the values of the parameters of the closure
            that are in calibration.

        Returns
        -------
        loss : float, positive
            Value of the loss function for the :code:`x` values of the closure
            parameters.
        """
        closure_parameters = self.coef_fit_params.fit_to_closure(x)
        scm_set = []
        for model in self.model_list:
            traj = model.compute_trajectory_with(closure_parameters)
            scm_set.append(traj)
        return self.loss(scm_set, self.database)


    def __call__(self):
        """
        Execute the callibration.

        First the optimizer is selected and parametrized with optax and the
        gradient function of the loss is computed. Then in the calibration
        loop, the gradient is evaluated on the currents values of the closure
        parameters, the eventual output are computed and the optax optimizer
        is updated. The optmizer used is :func:`optax.adam`.

        Returns
        -------
        closure_params : ClosureParametersAbstract
            The instance of the closure parameters changed with the final
            value of the calibrated parameters.
        """
        optimizer = optax.adam(self.learning_rate)
        x = self.coef_fit_params.gen_init_val()
        opt_state = optimizer.init(x)
        grad_loss = grad(self.loss_wrapped)
        for i in range(self.nloop):
            # compute the gradient
            grads = grad_loss(x)
            # print evolution
            if self.verbatim:
                print(f"""
                    loop {i}
                    x {x}
                    grads {grads}
                """)
            # write evolution
            if self.output_path is not None:
                data = np.load(self.output_path, allow_pickle=True)
                x_ev = data['x']
                grads_ev = data['grads']
                loss_it = data['loss_it']
                loss_values = data['loss_values']
                x_ev = np.hstack([x_ev, x.reshape(-1, 1)])
                grads_ev = np.hstack([grads_ev, grads.reshape(-1, 1)])
                if self.nit_loss != -1 and i%self.nit_loss == 0:
                    loss_it = np.append(loss_it, i)
                    loss_values = np.append(loss_values, self.loss_wrapped(x))
                np.savez(
                    self.output_path, x=x_ev, grads=grads_ev, loss_it=loss_it,
                    loss_values=loss_values
                )
            # update the optimizer
            updates, opt_state = optimizer.update(grads, opt_state)
            x = optax.apply_updates(x, updates)
        return self.coef_fit_params.fit_to_closure(x)



"""
Abstraction for calibration databases.


This module the objects that are used in Tunax to describe a :class:`Database`
of observations used for a calibration. By *obersvations* (:class:`Obs`), we
refer to a set of time-series representing a physical experiment, a measurment
or a simulation like a Large Eddy Simulation (LES) for example. These classes
can be obtained by the prefix :code:`tunax.database.` or directly by
:code:`tunax.`.

"""

from __future__ import annotations
from typing import List, Dict

import yaml
import xarray as xr
import equinox as eqx
import jax.numpy as jnp

from tunax.space import Grid, Trajectory
from tunax.case import Case


class Obs(eqx.Module):
    """
    Abstraction to represent an *obersation*.

    The *observations* represent every elements of a database, each one
    represent a simulation or a measurement with their own time-series of
    variables and the physical case which is linked to them

    Parameters
    ----------
    trajectory : Trajectory
        cf. attribute.
    case : Case
        cf. attribute.

    Attributes
    ----------
    trajectory : Trajectory
        The time-series of the variables that represent this obervation.
    case : Case
        The physical case that represent this observation.

    Raises
    ------
    ValueError
        If the :attr:`~space.Trajectory.time` of :attr:`trajectory` is not
        build with constant time-steps.

    """

    trajectory: Trajectory
    case: Case

    def __init__(self, trajectory: Trajectory, case: Case):
        time = trajectory.time
        steps = time[1:] - time[:-1]
        if not jnp.all(steps == steps[0]):
            raise ValueError('Tunax only handle constant output time-steps')
        self.trajectory = trajectory
        self.case = case

    @classmethod
    def from_files(
            cls,
            nc_path: str,
            yaml_path: str,
            var_names: Dict[str, str]
        ) -> Obs:
        """
        Create an instance from a *netcdf* and a *yaml* files.

        This class method build a trajectory from the :code:`.nc` file
        :code:`nc_path`, it build the physical parameters from the
        configuration file :code:`yaml_path`. :code:`var_names` is used to do
        the link between Tunax name convention and the one from the used
        database.

        Parameters
        ----------
        nc_path : str
            Path of the *netcdf* file that contains the time-series of the
            observation trajectory. The file should contains at least the
            three dimensions :attr:`~space.Grid.zr` :attr:`~space.Grid.zw` and
            :attr:`~space.Trajectory.time`. The time-series can be created with
            default values if they are not present in the file. Otherwise, they
            must have the good dimensions described in
            :class:`~space.Trajectory`.
        yaml_path : str
            Path of the *yaml* file that contains the parameters and forcing
            that describe the observation. The parameters should be float
            numbers and directly accessible from the root of the file with
            a key. Only the parameters that are described in
            :class:`~case.Case` will be takend in account.
        var_names : Dict[str, str]
            Link between the convention names in Tunax and the ones in the
            database. The keys are the Tunax names and the values are the names
            in the database. It works for variables of the
            :class:`~space.Trajectory` and fornthe parameters of
            :class:`~case.Case`. It must at least contains entries for
            :attr:`~space.Grid.zr` :attr:`~space.Grid.zw` and
            :attr:`~space.Trajectory.time`
        
        Returns
        -------
        obs : Obs
            An object that represent these files as an observation.
        """
        ds = xr.load_dataset(nc_path)
        # dimensions
        zr = jnp.array(ds[var_names['zr']].values)
        zw = jnp.array(ds[var_names['zw']].values)
        grid = Grid(zr, zw)
        time = jnp.array(ds[var_names['time']].values)
        nt, = time.shape
        nz = grid.nz
        # variables
        t_name = var_names['t']
        if t_name == '':
            t = jnp.full((nt, nz), 21.)
        else:
            t = jnp.array(ds[var_names['t']].values)
        s_name = var_names['s']
        if s_name == '':
            s = jnp.full((nt, nz), 35.)
        else:
            s = jnp.array(ds[var_names['s']].values)
        u_name = var_names['u']
        if u_name == '':
            u = jnp.full((nt, nz), 0.)
        else:
            u = jnp.array(ds[var_names['u']].values)
        v_name = var_names['v']
        if v_name == '':
            v = jnp.full((nt, nz), 0.)
        else:
            v = jnp.array(ds[var_names['v']].values)
        # writing trajectory
        trajectory = Trajectory(grid, time, t, s, u, v)

        with open(yaml_path, 'r', encoding='utf-8') as f:
            metadatas = yaml.safe_load(f)

        case = Case()
        case_attributes = list(vars(case).keys())
        for att in case_attributes:
            if att in var_names.keys():
                case = eqx.tree_at(
                    lambda t: getattr(t, att), case,
                    metadatas[var_names[att]])

        return cls(trajectory, case)


class Database(eqx.Module):
    """
    Represent a set of several observations that form a database.

    Parameters
    ----------
    observations : List[Obs]
        cf. attribute.

    Attributes
    ----------
    observations : List[Obs]
        A list of several observations with potentially various forcings,
        geometry and time configuration.


    """

    observations: List[Obs]
